import functools
import io
import os
import re
import socket
from unittest import mock
from urllib.parse import urlparse
import warnings

import numpy as np


try:
    import requests
except ModuleNotFoundError:
    requests = mock.Mock()
    REQUESTS_AVAILABLE = False
else:
    REQUESTS_AVAILABLE = True


#: Regular expression for matching a regular HTTP URL
REGEXP_HTTP_URL = re.compile(
    r"^(https?:\/\/)"  # protocol (http or https or omitted)
    r"([a-z0-9-\.]*)(\:[0-9]*)?\/"  # host:port
    r".+"  # path
)


class ETagNotInResponseHeaderWarning(UserWarning):
    """Used for cases where the requests.Response does not contain an ETag"""


class ConnectionTimeoutWarning(UserWarning):
    """Used when a connection fails or times out"""


class HTTPFile(io.IOBase):
    def __init__(self, url, chunk_size=2**18, keep_chunks=200):
        """Chunk-cached access to a URL supporting range requests

        Range requests (https://en.wikipedia.org/wiki/Byte_serving)
        allow clients to access specific parts of a file via HTTP
        without downloading the entire file.

        This class creates a file-like object from a URL that can
        then be passed on to e.g. h5py for reading. In addition, this
        class keeps a chunk cache of the URL, making it (A) fast to
        access frequently used parts of the file and (B) fast to slice
        through large files since the ratio of data downloaded versus
        (time-consuming) HTTP requests is very large.

        Parameters
        ----------
        url: str
            Path to the URL
        chunk_size: int
            Download chunk size. The entire file is split up into
            equally-sized (and thus indexable) chunks.
        keep_chunks: int
            Number of downloaded chunks to keep in memory. For a
            `chunk_size` of 2**18 bytes, a `keep_chunks` of 200
            impliese a chunk cache size of 50 MiB.
        """
        self.url = url
        self._chunk_size = chunk_size
        self._keep_chunks = keep_chunks
        self.session = session_cache.get_session(url)
        self._len = None
        self._etag = None
        self._pos = 0
        self.cache = {}

    def _parse_header(self):
        """parse the header sent by the server, populates length and etag"""
        if self._len is None:
            # Do not use `self.session.head`, because it might return a
            # wrong content-length for pre-signed S3 URLS.
            resp = self.session.get(self.url, stream=True, timeout=0.5)
            if resp.status_code != 200:
                raise ValueError(
                    f"Server replied with status code {resp.status_code} "
                    f"{resp.reason} for '{self.url}'")
            self._len = int(resp.headers["content-length"])
            # Try to determine the etag of the file.
            etag = resp.headers.get("etag", "").strip("'").strip('"')
            if len(etag) < 5:
                etag = None
                warnings.warn(f"Got empty ETag header for {self.url}",
                              ETagNotInResponseHeaderWarning)
            self._etag = etag

    @property
    def etag(self):
        """Unique identifier for this resource (version) by the web server"""
        self._parse_header()
        return self._etag

    @property
    def length(self):
        self._parse_header()
        return self._len

    @property
    def max_cache_size(self):
        """The maximum cache size allowed by `chunk_size` and `keep_chunks`"""
        return self._chunk_size * self._keep_chunks

    def close(self):
        """Close the file

        This closes the requests session and then calls `close` on
        the super class.
        """
        self.session.close()
        super(HTTPFile, self).close()

    def download_range(self, start, stop):
        """Download bytes given by the range (`start`, `stop`)

        `stop` is not inclusive (In the HTTP range request it normally is).
        """
        resp = self.session.get(self.url,
                                headers={"Range": f"bytes={start}-{stop-1}"}
                                )
        return resp.content

    def get_cache_chunk(self, index):
        """Return the cache chunk defined by `index`

        If the chunk is not in `self.cache`, it is downloaded.
        """
        if index not in self.cache:
            start = index*self._chunk_size
            stop = min((index+1)*self._chunk_size, self.length)
            self.cache[index] = self.download_range(start, stop)
        if len(self.cache) > self._keep_chunks:
            for kk in self.cache.keys():
                if kk != 0:  # always keep the first chunk
                    self.cache.pop(kk)
                    break
        return self.cache[index]

    def read(self, size=-1, /):
        """Cache-supported read operation (file object)"""
        data = self.read_range_cached(self._pos, self._pos + size)
        if size > 0:
            self._pos += size
        else:
            self._pos = self.length
        return data

    def read_range_cached(self, start, stop):
        """Concatenate the requested bytes from the cached chunks

        This calls `get_cache_chunk` and thus downloads cache
        chunks when necessary.
        """
        toread = stop - start
        # compute the chunk indices between start and stop
        chunk_start = np.int64(start // self._chunk_size)
        chunk_stop = np.int64(stop // self._chunk_size + 1)
        data = b""
        pos = start
        for chunk_index in range(chunk_start, chunk_stop):
            chunk = self.get_cache_chunk(chunk_index)
            chunk_start = pos % self._chunk_size
            if toread == 0:
                break
            elif chunk_start + toread >= self._chunk_size:
                data += chunk[chunk_start:]
                chunks_read = self._chunk_size - chunk_start
            else:
                chunk_end = stop % self._chunk_size
                data += chunk[chunk_start:chunk_end]
                chunks_read = chunk_end - chunk_start
            toread -= chunks_read
            pos += chunks_read

        return data

    def seek(self, offset, whence=os.SEEK_SET):
        """Seek to a position (file object)"""
        if whence == os.SEEK_SET:
            self._pos = offset
        elif whence == os.SEEK_CUR:
            self._pos += offset
        elif whence == os.SEEK_END:
            self._pos = self.length + offset

    def seekable(self):
        """The HTTP file is seekable"""
        return True

    def tell(self):
        """Tell the position (file object)"""
        return self._pos


class ResoluteRequestsSessionCache:
    def __init__(self):
        """A multiprocessing-safe cache for requests session objects

        This class implements empty `__getstate__` and `__setstate__`
        methods, so that when used in a multiprocessing context, sessions
        are never mirrored to the subprocesses. Each subprocess creates
        its own sessions.

        Note that only :class:`ResoluteRequestsSession` objects are used,
        which is ideal for the use-case of unstable internet connections.
        """
        #: This dictionary holds all sessions in use by the current process.
        #: Sessions are stored with the host name / netloc as the key.
        self.sessions = {}

    def __getstate__(self):
        """Returns None, so sessions are not pickled into subrpocesses"""
        pass

    def __setstate__(self, state):
        """Does nothing (see `__getstate__`)"""
        pass

    def get_session(self, url: str):
        """Return a requests session for the specified URL

        For each hostname, a different session is returned,
        but for identical hostnames, cached sessions are used.
        """
        urlp = urlparse(url)
        key = urlp.netloc
        if key not in self.sessions:
            self.sessions[key] = ResoluteRequestsSession()
        return self.sessions[key]


class ResoluteRequestsSession(requests.Session):
    """A session with built-in retry for `get`"""
    def get(self, *args, **kwargs):
        kwargs.setdefault("timeout", 0.5)
        for ii in range(100):
            try:
                resp = super(ResoluteRequestsSession,
                             self).get(*args, **kwargs)
            except (requests.exceptions.ConnectionError,
                    requests.exceptions.ReadTimeout,
                    requests.exceptions.ConnectTimeout,
                    requests.urllib3.exceptions.ConnectionError,
                    requests.urllib3.exceptions.ReadTimeoutError) as e:
                warnings.warn(f"Encountered {e} for {args} {kwargs}",
                              ConnectionTimeoutWarning)
                continue
            else:
                break
        else:
            raise requests.exceptions.ReadTimeout(
                f"Resolut sesion failed for {args} and {kwargs}!")
        return resp


def is_url_available(url: str, ret_reason=False):
    """Check whether a URL is available

    Parameters
    ----------
    url: str
        full URL to the object
    ret_reason: bool
        whether to return reason for unavailability

    Returns
    -------
    available: bool
        whether the URL is available
    reason: str
        reason for the URL not being available is `available` is False
    """
    avail = False
    reason = "none"
    if is_http_url(url):
        urlp = urlparse(url)
        # default to https if no scheme or port is specified
        port = urlp.port or (80 if urlp.scheme == "http" else 443)
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.settimeout(1)
            # Try to connect to the host
            try:
                # Use `hostname`, not `netloc`, because `netloc` contains
                # the port number which we do not want here.
                s.connect((urlp.hostname, port))
            except (socket.gaierror, OSError):
                reason = "no connection"
            else:
                # Try to access the url
                try:
                    ses = session_cache.get_session(url)
                    req = ses.get(url, stream=True, timeout=1)
                    avail = req.ok
                    if not avail:
                        reason = req.reason.lower()
                except OSError:
                    reason = "oserror"
                    pass
    else:
        reason = "invalid"
    if ret_reason:
        return avail, reason
    else:
        return avail


@functools.lru_cache()
def is_http_url(string):
    """Check whether `string` is a valid URL using regexp"""
    if not isinstance(string, str):
        return False
    else:
        return REGEXP_HTTP_URL.match(string.strip())


#: cache of requests sessions for current process
session_cache = ResoluteRequestsSessionCache()



"""Cache for fast "recomputation"
"""
import functools
import gc
import hashlib

import numpy as np


MAX_SIZE = 100


class Cache:
    _cache = {}
    _keys = []

    def __init__(self, func):
        """
        A cache that can be used to decorate methods that accept
        numpy ndarrays as arguments.

        - cache is based on dictionary
        - md5 hashes of method arguments are the dictionary keys
        - applicable decorator for all methods in a module
        - applicable to methods with the same name in different
          source files
        - set cache size with `cached.MAX_SIZE`
        - only one global cache is generated, there are no instances
          of `Cache`

        Notes
        -----
        If you are using other decorators with this decorator, please
        make sure to apply the `Cache` first (first line before method
        definition). This wrapper uses name, doc, and filename of the
        method to identify it. If another wrapper does not implement
        a unique `__doc__` and is applied to multiple methods, then
        `Cached` might return values of the wrong method.
        """
        self.func = func
        functools.update_wrapper(self, func)

    def __call__(self, *args, **kwargs):
        self.ahash = hashlib.md5()

        # hash arguments
        for arg in args:
            self._update_hash(arg)

        # hash keyword arguments
        kwds = list(kwargs.keys())
        kwds.sort()
        for k in kwds:
            self._update_hash(k)
            self._update_hash(kwargs[k])

        # make sure we are caching for the correct method
        self._update_hash(self.func.__name__)
        self._update_hash(self.func.__doc__)
        self._update_hash(self.func.__code__.co_filename)

        ref = self.ahash.hexdigest()

        if ref in Cache._cache:
            return Cache._cache[ref]
        else:
            data = self.func(*args, **kwargs)
            Cache._cache[ref] = data
            Cache._keys.append(ref)
            if len(Cache._keys) > MAX_SIZE:
                delref = Cache._keys.pop(0)
                Cache._cache.pop(delref)
            return data

    def _update_hash(self, arg):
        """Takes an argument and updates the hash.
        The argument can be an np.array, string, or list
        of things that are convertable to strings.
        """
        if isinstance(arg, np.ndarray):
            self.ahash.update(arg.view(np.uint8))
        elif isinstance(arg, list):
            [self._update_hash(a) for a in arg]
        else:
            self.ahash.update(str(arg).encode('utf-8'))

    @staticmethod
    def clear_cache():
        """Remove all cached objects"""
        del Cache._keys
        for k in list(Cache._cache.keys()):
            it = Cache._cache.pop(k)
            del it
        del Cache._cache
        Cache._keys = []
        Cache._cache = {}
        gc.collect()



from dclab import cli, new_dataset, rtdc_dataset, RTDCWriter

import h5py
import numpy as np
import pytest

from helper_methods import retrieve_data


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_repack_basic():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("repacked.rtdc")

    ret = cli.repack(path_in=path_in, path_out=path_out)
    assert ret is None, "by default, this method should return 0 (exit 0)"

    with new_dataset(path_out) as dsj, new_dataset(path_in) as ds0:
        assert len(dsj)
        assert len(dsj) == len(ds0)
        for feat in ds0.features_innate:
            if feat in ds0.features_scalar:
                assert np.all(dsj[feat] == ds0[feat]), feat
        for ii in range(len(ds0)):
            assert np.all(dsj["contour"][ii] == ds0["contour"][ii])
            assert np.all(dsj["image"][ii] == ds0["image"][ii])
            assert np.all(dsj["mask"][ii] == ds0["mask"][ii])


@pytest.mark.parametrize("use_basins", [True, False])
def test_repack_basin_internal(use_basins):
    """
    Internal basins should just be copied to the new file
    """
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")
    h5path_out = h5path.with_name("repacked.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        hw.store_basin(basin_name="example basin",
                       basin_type="internal",
                       basin_format="h5dataset",
                       basin_locs=["basin_events"],
                       basin_descr="an example test basin",
                       internal_data={"userdef1": np.arange(2)},
                       basin_map=np.zeros(src["events/deform"].shape[0]),
                       basin_feats=["userdef1"],
                       )

    # sanity check
    with new_dataset(h5path_small) as ds:
        assert "userdef1" in ds.features_basin
        assert "userdef1" not in ds.features_innate

    # compress the basin-based dataset
    cli.repack(path_in=h5path_small, path_out=h5path_out,
               strip_basins=not use_basins)

    with h5py.File(h5path_out) as h5:
        assert "deform" in h5["events"], "sanity check"
        # userdef1 should not be in "events" in any case
        assert "userdef1" not in h5["events"]
        if use_basins:
            assert "userdef1" in h5["basin_events"]
            assert np.all(h5["basin_events"]["userdef1"] == np.arange(2))
        else:
            assert "basin_events" not in h5

    with new_dataset(h5path_out) as ds:
        if use_basins:
            assert "userdef1" in ds.features_basin
        else:
            assert "userdef1" not in ds.features_basin


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_repack_multiple_basins():
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")
    h5path_out = h5path.with_name("repacked.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        hw.store_basin(basin_name="example basin",
                       basin_type="file",
                       basin_format="hdf5",
                       basin_locs=[h5path],
                       basin_descr="an example test basin",
                       )

        hw.store_basin(basin_name="example basin",
                       basin_type="file",
                       basin_format="hdf5",
                       basin_locs=[h5path],
                       basin_descr="a different test basin",
                       )

        assert len(hw.h5file["basins"].keys()) == 2

    # repack the basin-based dataset
    cli.repack(path_in=h5path_small,
               path_out=h5path_out)

    with new_dataset(h5path_out) as ds:
        assert len(ds.basins) == 2


def test_repack_no_data_from_basins_written():
    """
    When repacking a dataset, feature data from the basin should not be
    written to the output file.
    """
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")
    h5path_out = h5path.with_name("repacked.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        hw.store_basin(basin_name="example basin",
                       basin_type="file",
                       basin_format="hdf5",
                       basin_locs=[h5path],
                       basin_descr="an example test basin",
                       )

    # sanity check
    with new_dataset(h5path_small) as ds:
        assert "image" in ds.features

    # repack the basin-based dataset
    cli.repack(path_in=h5path_small, path_out=h5path_out)

    with h5py.File(h5path_out) as h5:
        assert "deform" in h5["events"], "sanity check"
        assert "image" not in h5["events"], "Arrgh, basin feature was copied"

    with new_dataset(h5path_out) as ds:
        assert "image" in ds.features_basin


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_repack_remove_secrets():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("repacked.rtdc")

    with h5py.File(path_in, "a") as h5:
        h5.attrs["experiment:sample"] = "my dirty secret"

    with h5py.File(path_in, "a") as h5:
        h5.attrs["experiment:sample"] = "sunshine"

    # test whether the dirty secret is still there
    with open(str(path_in), "rb") as fd:
        data = fd.read()
        assert str(data).count("my dirty secret")

    # now repack
    cli.repack(path_in=path_in, path_out=path_out)

    # clean?
    with open(str(path_out), "rb") as fd:
        data = fd.read()
        assert not str(data).count("my dirty secret")


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_repack_strip_logs():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("repacked.rtdc")

    # write some logs
    with h5py.File(path_in, "a") as h5:
        hw = rtdc_dataset.RTDCWriter(h5)
        hw.store_log("test_log", ["peter", "hans"])

    cli.repack(path_in=path_in, path_out=path_out, strip_logs=True)

    with new_dataset(path_out) as dsj, new_dataset(path_in) as ds0:
        assert ds0.logs
        assert not dsj.logs


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_repack_strip_basins():
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")
    h5path_out = h5path.with_name("repacked.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        hw.store_basin(basin_name="example basin",
                       basin_type="file",
                       basin_format="hdf5",
                       basin_locs=[h5path],
                       basin_descr="an example test basin",
                       )

    # sanity check
    with new_dataset(h5path_small) as ds:
        assert "image" in ds.features

    # repack the basin-based dataset
    cli.repack(path_in=h5path_small,
               path_out=h5path_out,
               strip_basins=True)

    with h5py.File(h5path_out) as h5:
        assert "deform" in h5["events"], "sanity check"
        assert "image" not in h5["events"], "Arrgh, basin feature was copied"

    with new_dataset(h5path_out) as ds:
        assert "image" not in ds.features_basin, "[sic] we wanted to strip"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_repack_user_metadata():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    with h5py.File(path_in, "a") as h5:
        h5.attrs["user:peter"] = "hans"

    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("repacked.rtdc")

    cli.repack(path_in=path_in, path_out=path_out)

    with new_dataset(path_out) as ds:
        assert ds.config["user"]["peter"] == "hans"



"""RT-DC hdf5 format"""
from __future__ import annotations

from ...external.packaging import parse as parse_version


def get_software_version_from_h5(h5):
    software_version = h5.attrs.get("setup:software version", "")
    if isinstance(software_version, bytes):
        software_version = software_version.decode("utf-8")
    return software_version


def is_defective_feature_aspect(h5):
    """In Shape-In 2.0.6, there was a wrong variable cast"""
    software_version = get_software_version_from_h5(h5)
    return software_version in ["ShapeIn 2.0.6", "ShapeIn 2.0.7"]


def is_defective_feature_time(h5):
    """Shape-In stores the "time" feature as a low-precision float32

    This makes time resolution for large measurements useless,
    because times are only resolved with four digits after the
    decimal point. Here, we first check whether the "frame" feature
    and the [imaging]:"frame rate" configuration are set. If so,
    then we can compute "time" as an ancillary feature which will
    be more accurate than its float32 version.
    """
    # This is a necessary requirement. If we cannot compute the
    # ancillary feature, then we cannot ignore (even inaccurate) information.
    has_ancil = "frame" in h5["events"] and h5.attrs.get("imaging:frame rate",
                                                         0) != 0
    if not has_ancil:
        return False

    # If we have a 32 bit dataset, then things are pretty clear.
    is_32float = h5["events/time"].dtype.char[-1] == "f"
    if is_32float:
        return True

    # Consider the software
    software_version = get_software_version_from_h5(h5)

    # Only Shape-In stores false data, so we can ignore other recording
    # software.
    is_shapein = software_version.count("ShapeIn")
    if not is_shapein:
        return False

    # The tricky part: dclab might have analyzed the dataset recorded by
    # Shape-In, e.g. in a compression step. Since dclab appends its version
    # string to the software_version, we just have to parse that and make
    # sure that it is above 0.47.6.
    last_version = software_version.split("|")[-1].strip()
    if last_version.startswith("dclab"):
        dclab_version = last_version.split()[1]
        if parse_version(dclab_version) < parse_version("0.47.6"):
            # written with an older version of dclab
            return True

    # We covered all cases:
    # - ancillary information are available
    # - it's not a float32 dataset
    # - we excluded all non-Shape-In recording software
    # - it was not written with an older version of dclab
    return False


def is_defective_feature_volume(h5):
    """dclab computed volume wrong up until version 0.36.1"""
    # first check if the scripted fix was applied
    if "dclab_issue_141" in list(h5.get("logs", {}).keys()):
        return False
    # if that does not apply, check the software version
    software_version = get_software_version_from_h5(h5)
    if software_version:
        last_version = software_version.split("|")[-1].strip()
        if last_version.startswith("dclab"):
            dclab_version = last_version.split()[1]
            if parse_version(dclab_version) < parse_version("0.37.0"):
                return True
    return False


def is_defective_feature_inert_ratio(h5):
    """For long channels, there was an integer overflow until 0.48.1

    The problem here is that not only the channel length, but also
    the length of the contour play a role. All point coordinates of
    the contour are summed up and multiplied with one another which
    leads to integer overflows when computing mu20.

    Thus, this test is only a best guess, but still quite fast.

    See also https://github.com/DC-analysis/dclab/issues/212
    """
    # determine whether the image width is larger than 500px
    # If this file was written with dclab, then we always have the ROI size,
    # so we don't have to check the actual image.
    width_large = h5.attrs.get("imaging:roi size x", 0) > 500

    if width_large:
        # determine whether the software version was outdated
        software_version = get_software_version_from_h5(h5)
        if software_version:
            version_pipeline = [v.strip() for v in software_version.split("|")]
            last_version = version_pipeline[-1]
            if last_version.startswith("dclab"):
                dclab_version = last_version.split()[1]
                # The fix was implemented in 0.48.2, but this method here
                # was only implemented in 0.48.3, so we might have leaked
                # old data into new files.
                if parse_version(dclab_version) < parse_version("0.48.3"):
                    return True
    return False


def is_defective_feature_inert_ratio_raw_cvx(h5):
    """Additional check for `inert_ratio_raw` and `inert_ratio_cvx`

    These features were computed with Shape-In and were very likely
    computed correctly.

    See https://github.com/DC-analysis/dclab/issues/224
    """
    if is_defective_feature_inert_ratio(h5):
        # Possibly affected. Only return False if Shape-In check is negative
        software_version = get_software_version_from_h5(h5)
        version_pipeline = [v.strip() for v in software_version.split("|")]
        first_version = version_pipeline[0]
        if first_version.startswith("ShapeIn"):
            si_version = first_version.split()[1]
        elif "shapein-acquisition" in h5.get("logs", []):
            # Later versions of Shape-In do not anymore write "ShapeIn" in the
            # version string.
            si_version = first_version
        else:
            # Some other software was used to record the data and dclab
            # very likely stored the wrong inertia ratio.
            return True

        # We trust Shape-In >= 2.0.5
        if parse_version(si_version) >= parse_version("2.0.5"):
            return False

        return True

    return False


#: dictionary of defective features, defined by HDF5 attributes;
#: if a value matches the given HDF5 attribute, the feature is
#: considered defective
DEFECTIVE_FEATURES = {
    # feature: [HDF5_attribute, matching_value]
    "aspect": is_defective_feature_aspect,
    "inert_ratio_cvx": is_defective_feature_inert_ratio_raw_cvx,
    "inert_ratio_prnc": is_defective_feature_inert_ratio,
    "inert_ratio_raw": is_defective_feature_inert_ratio_raw_cvx,
    "tilt": is_defective_feature_inert_ratio,
    "time": is_defective_feature_time,
    "volume": is_defective_feature_volume,
}



"""
Testing script for volume.py
An ellipse is created and the analytical and numerical solution are compared
"""
import itertools

import numpy as np
import pytest

import dclab
from dclab.features.volume import get_volume, vol_revolve

from helper_methods import retrieve_data


def area_of_polygon(x, y):
    return 0.5 * np.abs(np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))


def centroid_of_polygon(points):
    """
    http://stackoverflow.com/a/14115494/190597 (mgamba)
    Centroid of polygon:
    http://en.wikipedia.org/wiki/Centroid#Centroid_of_polygon
    """
    area = area_of_polygon(*zip(*points))
    result_x = 0
    result_y = 0
    N = len(points)
    points = itertools.cycle(points)
    x1, y1 = next(points)
    for _i in range(N):
        x0, y0 = x1, y1
        x1, y1 = next(points)
        cross = (x0 * y1) - (x1 * y0)
        result_x += (x0 + x1) * cross
        result_y += (y0 + y1) * cross
    result_x /= (area * 6.0)
    result_y /= (area * 6.0)
    return abs(result_x), abs(result_y)


def get_ellipse_coords(a, b, x=0.0, y=0.0, angle=0.0, k=2):
    """

    Draws an ellipse using (360*k + 1) discrete points; based on pseudo code
    given at http://en.wikipedia.org/wiki/Ellipse

    k = 1 means 361 points (degree by degree)
    a = major axis distance,
    b = minor axis distance,
    x = offset along the x-axis
    y = offset along the y-axis
    angle = clockwise rotation [in degrees] of the ellipse;
        * angle=0  : the ellipse is aligned with the positive x-axis
        * angle=30 : rotated 30 degrees clockwise from positive x-axis
    """
    pts = np.zeros((360 * k + 1, 2))

    beta = -angle * np.pi / 180
    sin_beta = np.sin(beta)
    cos_beta = np.cos(beta)
    alpha = np.linspace(0, 2*np.pi, 360 * k + 1, endpoint=True)

    sin_alpha = np.sin(alpha)
    cos_alpha = np.cos(alpha)

    pts[:, 0] = x + (a * cos_alpha * cos_beta - b * sin_alpha * sin_beta)
    pts[:, 1] = y + (a * cos_alpha * sin_beta + b * sin_alpha * cos_beta)

    return pts


def test_af_volume():
    pytest.importorskip("nptdms")
    ds = dclab.new_dataset(retrieve_data("fmt-tdms_minimal_2016.zip"))
    vol = ds["volume"]
    # There are a lot of nans, because the contour is not given everywhere
    vol = vol[~np.isnan(vol)]
    assert np.allclose(vol[0], 571.0512910465295)
    assert np.allclose(vol[12], 1008.5405681260276)


def test_get_volume():
    # Helper definitions to get an ellipse
    major = 10
    minor = 5
    ellip = get_ellipse_coords(a=major,
                               b=minor,
                               x=minor,
                               y=5,
                               angle=0,
                               k=100)
    # obtain the centroid (corresponds to pos_x and pos_lat)
    cx, cy = centroid_of_polygon(ellip)
    volume = get_volume(cont=[ellip],
                        pos_x=[cx],
                        pos_y=[cy],
                        pix=1,
                        fix_orientation=True)

    # Analytic solution for volume of ellipsoid:
    v = 4 / 3 * np.pi * major * minor**2
    msg = "Calculation of volume is faulty!"
    assert np.allclose(np.array(volume), np.array(v)), msg


def test_orientation():
    """counter-clockwise vs clockwise"""
    # Helper definitions to get an ellipse
    major = 10
    minor = 5
    ellip = get_ellipse_coords(a=major,
                               b=minor,
                               x=minor,
                               y=5,
                               angle=0,
                               k=100)
    # obtain the centroid (corresponds to pos_x and pos_lat)
    cx, cy = centroid_of_polygon(ellip)

    v1 = get_volume(cont=ellip,
                    pos_x=cx,
                    pos_y=cy,
                    pix=1,
                    fix_orientation=True)

    # Turn contour around
    v2 = get_volume(cont=ellip[::-1, :],
                    pos_x=cx,
                    pos_y=cy,
                    pix=1,
                    fix_orientation=True)

    assert np.allclose(v1, v2, atol=1e-11, rtol=0)


def test_shape():
    major = 10
    minor = 5
    ellip = get_ellipse_coords(a=major,
                               b=minor,
                               x=minor,
                               y=5.0,
                               angle=0,
                               k=100)
    cx, cy = centroid_of_polygon(ellip)
    # no lists
    volume = get_volume(cont=ellip,
                        pos_x=cx,
                        pos_y=cy,
                        pix=1)
    assert isinstance(volume, float)

    volumelist = get_volume(cont=[ellip],
                            pos_x=[cx],
                            pos_y=[cy],
                            pix=1)
    assert isinstance(volumelist, np.ndarray)


def test_xpos():
    """xpos is not necessary to compute volume dense ellipse"""
    major = 10
    minor = 5
    ellip = get_ellipse_coords(a=major,
                               b=minor,
                               x=minor,
                               y=5.0,
                               angle=0,
                               k=100)
    cx, cy = centroid_of_polygon(ellip)
    # no lists
    v0 = get_volume(cont=ellip,
                    pos_x=cx,
                    pos_y=cy,
                    pix=1)
    for cxi in np.linspace(0, 2 * cx, 10):
        vi = get_volume(cont=ellip,
                        pos_x=cxi,
                        pos_y=cy,
                        pix=1)
        assert np.allclose(v0, vi)


@pytest.mark.parametrize("npoints,rtol", [[100, 6.72e-4],
                                          [1000, 6.6e-6]])
def test_vol_revolve_circular_toroid(npoints, rtol):
    """Upstream test function 1

    https://de.mathworks.com/matlabcentral/fileexchange/36525-volrevolve

    % Verification for circular toroid of major radius R0, minor radius a
    % Volume is 2 * pi^2 * R0 * a^2. Run this code:
    % clear all
    % R0 = 5 ;
    % a = 1 ;
    % npoints = 100 ;
    % theta = 2*pi*[0:1:npoints-1]'/double(npoints-1) ;
    % R = R0 + a*cos(theta) ;
    % Z =      a*sin(theta) ;
    % vol_analytic = 2 * pi^2 * R0 * a^2 ;
    %  >> 98.6960
    % vol = volRevolve(R,Z) ;
    %  >> 98.6298 (6.7e-04 relative error)
    % Do it again with npoints = 1000, get:
    %  >> 98.6954 (6.6e-06 relative error)
    % As expected, it's always slightly small because the polygon inscribes the
    % circle.
    """
    r0 = 5
    a = 1
    theta = 2 * np.pi * np.arange(npoints-1) / (npoints-1)
    r = r0 + a*np.cos(theta)
    z = a*np.sin(theta)
    vol_analytic = 2 * np.pi**2 * r0 * a**2
    vol = vol_revolve(r, z)
    assert np.allclose(vol_analytic, 98.6960, rtol=0, atol=0.001)
    assert np.allclose(vol_analytic, vol, rtol=rtol, atol=0)
    assert vol < vol_analytic


def test_vol_revolve_rectangular_toroid():
    """Upstream test function 2

    https://de.mathworks.com/matlabcentral/fileexchange/36525-volrevolve

    % Verification for washer (rectangular toroid), with the radius of the
    % 'hole' in the washer being a, and the outer radius of the washer being b.
    % (Thus the width of the metal cross section is b-a.) The height of the
    % washer is h. Then the volume is pi * (b^2 - a^2) * h. Run this code:
    clear all
    a = 1 ;
    b = 2 ;
    h = 10 ;
    R = [a; b; b; a; a] ;
    Z = [0; 0; h; h; 0] ;
    vol_analytic = pi * (b^2 - a^2) * h ;
    % >> 94.2478
    vol = volRevolve(R,Z) ;
    % >> 94.2478
    """
    a = 1
    b = 2
    h = 10
    r = [a, b, b, a, a]
    z = [0, 0, h, h, 0]
    vol_analytic = np.pi * (b**2 - a**2) * h
    vol = vol_revolve(r, z)
    assert vol_analytic == vol
    assert np.allclose(vol, 94.2478, rtol=0, atol=0.0001)



# flake8: noqa: F401
from .base import RTDC_HDF5, MIN_DCLAB_EXPORT_VERSION
from .basin import HDF5Basin  # import means registering
from .events import (
    H5Events, H5MaskEvent, H5TraceEvent, H5ScalarEvent, H5ContourEvent)
from .feat_defect import DEFECTIVE_FEATURES



import hashlib
import json
import numbers
import pathlib
import platform
import time
import warnings

import h5py
import numpy as np

try:
    import imageio
except ModuleNotFoundError:
    imageio = None

try:
    import nptdms
except ModuleNotFoundError:
    nptdms = None

from ..rtdc_dataset import fmt_tdms
from .. import util
from .._version import version


class ExtendedJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, pathlib.Path):
            return str(obj)
        elif isinstance(obj, numbers.Integral):
            return int(obj)
        elif isinstance(obj, np.bool_):
            return bool(obj)
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


def assemble_warnings(w):
    """pretty-format all warnings for logs"""
    wlog = []
    for wi in w:
        wlog.append(f"{wi.category.__name__}")
        wlog.append(f" in {wi.category.__module__} line {wi.lineno}:")
        # make sure line is not longer than 100 characters
        words = str(wi.message).split(" ")
        wline = "  "
        for ii in range(len(words)):
            wline += " " + words[ii]
            if ii == len(words) - 1:
                # end
                wlog.append(wline)
            elif len(wline + words[ii+1]) + 1 >= 100:
                # new line
                wlog.append(wline)
                wline = "  "
            # nothing to do here
    return wlog


def get_command_log(paths, custom_dict=None):
    """Return a json dump of system parameters

    Parameters
    ----------
    paths: list of pathlib.Path or str
        paths of related measurement files; up to 5MB of each of
        them is md5-hashed and included in the "files" key
    custom_dict: dict
        additional user-defined entries; must contain simple
        Python objects (json.dumps must still work)
    """
    if custom_dict is None:
        custom_dict = {}
    data = get_job_info()
    data["files"] = []
    for ii, pp in enumerate(paths):
        if pathlib.Path(pp).exists():
            fdict = {"name": pathlib.Path(pp).name,
                     # Hash only 5 MB of the input file
                     "md5-5M": util.hashfile(pp,
                                             count=80,
                                             constructor=hashlib.md5),
                     "index": ii+1
                     }
        else:
            fdict = {"name": f"{pp}",
                     "index": ii + 1
                     }
        data["files"].append(fdict)
    final_data = {}
    final_data.update(custom_dict)
    final_data.update(data)
    dump = json.dumps(final_data,
                      sort_keys=True,
                      indent=2,
                      cls=ExtendedJSONEncoder).split("\n")
    return dump


def get_job_info():
    """Return dictionary with current job information

    Returns
    -------
    info: dict of dicts
        Job information including details about time, system,
        python version, and libraries used.
    """
    data = {
        "utc": {
            "date": time.strftime("%Y-%m-%d", time.gmtime()),
            "time": time.strftime("%H:%M:%S", time.gmtime()),
        },
        "system": {
            "info": platform.platform(),
            "machine": platform.machine(),
            "name": platform.system(),
            "release": platform.release(),
            "version": platform.version(),
        },
        "python": {
            "build": ", ".join(platform.python_build()),
            "implementation": platform.python_implementation(),
            "version": platform.python_version(),
        },
        "libraries": {
            "dclab": version,
            "h5py": h5py.__version__,
            "numpy": np.__version__,
        }
    }
    if imageio is not None:
        data["libraries"]["imageio"] = imageio.__version__
    if nptdms is not None:
        data["libraries"]["nptdms"] = nptdms.__version__
    return data


def print_info(string):
    print(f"\033[1m{string}\033[0m")


def print_alert(string):
    print_info(f"\033[33m{string}")


def print_violation(string):
    print_info(f"\033[31m{string}")


def setup_task_paths(paths_in, paths_out, allowed_input_suffixes):
    """Setup directories for a CLI task

    Parameters
    ----------
    paths_in: list of str or lsit of pathlib.Path or str or pathlib.Path
        Input paths
    paths_out: list of str or list of pathlib.Path or str or pathlib.Path
        Output paths
    allowed_input_suffixes: list
        List of allowed input suffixes (e.g. [".rtdc"])

    Returns
    -------
    paths_in: list of pathlib.Path or pathlib.Path
        Input paths
    paths_out: list of pathlib.Path or pathlib.Path
        Output paths
    paths_temp: list of pathlib.Path or pathlib.Path
        Temporary paths (working path)
    """
    if isinstance(paths_in, list):
        list_in = True
    else:
        paths_in = [paths_in]
        list_in = False

    if isinstance(paths_out, list):
        list_out = True
    else:
        paths_out = [paths_out]
        list_out = False

    paths_in = [pathlib.Path(pi) for pi in paths_in]
    for pi in paths_in:
        if pi.suffix not in allowed_input_suffixes:
            raise ValueError(f"Unsupported file type: '{pi.suffix}'")

    paths_out = [pathlib.Path(po) for po in paths_out]
    for ii, po in enumerate(paths_out):
        if po.suffix != ".rtdc":
            paths_out[ii] = po.with_name(po.name + ".rtdc")
    [po.unlink() for po in paths_out if po.exists()]

    paths_temp = [po.with_suffix(".rtdc~") for po in paths_out]
    [pt.unlink() for pt in paths_temp if pt.exists()]

    # convert lists back to paths
    if not list_in:
        paths_in = paths_in[0]

    if not list_out:
        paths_out = paths_out[0]
        paths_temp = paths_temp[0]

    return paths_in, paths_out, paths_temp


def skip_empty_image_events(ds, initial=True, final=True):
    """Set a manual filter to skip inital or final empty image events"""
    if initial:
        if (("image" in ds and ds.format == "tdms"
             and ds.config["fmt_tdms"]["video frame offset"])
            or ("contour" in ds and np.all(ds["contour"][0] == 0))
                or ("image" in ds and np.all(ds["image"][0] == 0))):
            ds.filter.manual[0] = False
            ds.apply_filter()
    if final:
        # This is not easy to test, because we need a corrupt
        # frame.
        if "image" in ds:
            idfin = len(ds) - 1
            if ds.format == "tdms":
                with warnings.catch_warnings(record=True) as wfin:
                    warnings.simplefilter(
                        "always",
                        fmt_tdms.event_image.CorruptFrameWarning)
                    _ = ds["image"][idfin]  # provoke a warning
                for ww in wfin:
                    if ww.category == fmt_tdms.event_image.CorruptFrameWarning:
                        ds.filter.manual[idfin] = False
                        ds.apply_filter()
                        break
            elif np.all(ds["image"][idfin] == 0):
                ds.filter.manual[idfin] = False
                ds.apply_filter()



import json
import pathlib

import h5py
import numpy as np

import pytest

import dclab
from dclab import new_dataset, rtdc_dataset, RTDCWriter
from dclab.rtdc_dataset.fmt_hdf5.basin import HDF5Basin
from dclab.rtdc_dataset.feat_basin import BasinNotAvailableError


from helper_methods import retrieve_data


def test_basin_as_dict():
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        hw.store_basin(basin_name="example basin",
                       basin_type="file",
                       basin_format="hdf5",
                       basin_locs=[h5path],
                       basin_descr="an example test basin",
                       )

    with dclab.new_dataset(h5path_small) as ds:
        bdict = ds.basins[0].as_dict()
        assert bdict["basin_name"] == "example basin"
        assert bdict["basin_type"] == "file"
        assert bdict["basin_format"] == "hdf5"
        assert [str(p.resolve()) for p in bdict["basin_locs"]] == \
               [str(pathlib.Path(h5path).resolve())]
        assert bdict["basin_descr"] == "an example test basin"

    # Now use the data from `bdict` to create a new basin
    h5path_two = h5path.with_name("smaller_two.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_two) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        hw.store_basin(**bdict)

    del bdict
    del ds

    with dclab.new_dataset(h5path_small) as ds2:
        bdict2 = ds2.basins[0].as_dict()
        assert bdict2["basin_name"] == "example basin"
        assert bdict2["basin_type"] == "file"
        assert bdict2["basin_format"] == "hdf5"
        assert [str(p.resolve()) for p in bdict2["basin_locs"]] == \
               [str(pathlib.Path(h5path).resolve())]
        assert bdict2["basin_descr"] == "an example test basin"


def test_basin_not_available():
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, h5py.File(h5path_small, "w") as dst:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=dst,
                               features="scalar")
        # Next, store the basin information in the new dataset
        bdat = {
            "type": "file",
            "format": "hdf5",
            "paths": [
                "fake.rtdc",  # fake path
                str(h5path),  # absolute path name
            ]
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = dst.require_group("basins")
        with RTDCWriter(dst, mode="append") as hw:
            hw.write_text(basins, "my_basin", blines)
        # sanity checks
        assert "deform" in dst["events"]
        assert "image" not in dst["events"]

    h5path.unlink()

    # Now open the scalar dataset and check whether basins missing
    with new_dataset(h5path_small) as ds:
        assert "image" not in ds
        assert not ds.features_basin

    # Also test that on a lower level
    bn = HDF5Basin(h5path)
    assert not bn.is_available()
    with pytest.raises(BasinNotAvailableError, match="is not available"):
        _ = bn.ds


def test_basin_nothing_available():
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, h5py.File(h5path_small, "w") as dst:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=dst,
                               features="scalar")
        # Next, store the basin information in the new dataset
        bdat = {
            "type": "file",
            "format": "hdf5",
            "paths": [
                "fake.rtdc",  # fake path
            ]
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = dst.require_group("basins")
        with RTDCWriter(dst, mode="append") as hw:
            hw.write_text(basins, "my_basin", blines)
        # sanity checks
        assert "deform" in dst["events"]
        assert "image" not in dst["events"]

    h5path.unlink()

    # Now open the scalar dataset and check whether basins missing
    with new_dataset(h5path_small) as ds:
        assert "image" not in ds
        assert not ds.features_basin
        _ = ds["index"]


def test_basin_features_path_absolute():
    """Create a dataset that refers to a basin in an absolute path"""
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, h5py.File(h5path_small, "w") as dst:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=dst,
                               features="scalar")
        # Next, store the basin information in the new dataset
        bdat = {
            "type": "file",
            "format": "hdf5",
            "paths": [
                "fake.rtdc",  # fake path
                str(h5path),  # absolute path name
            ]
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = dst.require_group("basins")
        with RTDCWriter(dst, mode="append") as hw:
            hw.write_text(basins, "my_basin", blines)
        # sanity checks
        assert "deform" in dst["events"]
        assert "image" not in dst["events"]

    # Now open the scalar dataset and check whether basins are defined
    with new_dataset(h5path_small) as ds:
        assert "image" in ds.features_basin
        assert "image" not in ds.features_innate
        assert "image" in ds
        assert np.median(ds["image"][0]) == 151


def test_basin_features_path_relative():
    """Create a dataset that refers to a basin in a relative path"""
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, h5py.File(h5path_small, "w") as dst:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=dst,
                               features="scalar")
        # Next, store the basin information in the new dataset
        bdat = {
            "type": "file",
            "format": "hdf5",
            "paths": [
                "fake.rtdc",  # fake path
                h5path.name,  # relative path name
            ]
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = dst.require_group("basins")
        with RTDCWriter(dst, mode="append") as hw:
            hw.write_text(basins, "my_basin", blines)
        # sanity checks
        assert "deform" in dst["events"]
        assert "image" not in dst["events"]

    # Now open the scalar dataset and check whether basins are defined
    with new_dataset(h5path_small) as ds:
        assert "image" in ds.features_basin
        assert "image" not in ds.features_innate
        assert "image" in ds
        assert np.median(ds["image"][0]) == 151


def test_basin_identifiers_from_run_identifier():
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    with h5py.File(h5path, "a") as hw:
        hw.attrs["experiment:run identifier"] = "hey-you-ekelpack"
    with dclab.new_dataset(h5path) as ds:
        assert ds.get_measurement_identifier() == "hey-you-ekelpack"


def test_basin_identifiers_from_time_and_setup_id():
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    with dclab.new_dataset(h5path) as ds:
        assert ds.get_measurement_identifier() == \
               "01cd05a0-e084-0451-c470-2a2bf13c9e2d"


def test_basin_identifier_matches_file_none_invalid():
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    with h5py.File(h5path, "a") as h5:
        del h5.attrs["setup:identifier"]
    with dclab.new_dataset(h5path) as ds:
        assert not ds.get_measurement_identifier() == \
               "01cd05a0-e084-0451-c470-2a2bf13c9e2d"



class DCORLogs:
    def __init__(self, api):
        self.api = api
        self._logs_cache = None

    def __contains__(self, key):
        return key in self.keys()

    def __getitem__(self, key):
        return self._logs[key]

    def __iter__(self):
        for key in self.keys():
            yield key

    def __len__(self):
        return len(self._logs)

    def keys(self):
        return self._logs.keys()

    @property
    def _logs(self):
        if self._logs_cache is None:
            self._logs_cache = self.api.get(query="logs")
        return self._logs_cache



import numpy as np


def map_indices_child2parent(child, child_indices):
    """Map child RTDCBase event indices to parent RTDCBase

    Given a hierarchy child and indices defined for that child,
    return the corresponding indices for its parent.

    For instance, a child is defined in such a way that it
    has every second event of its parent (`parent.filter.all[::2]=False`
    i.e. the filtering array is `[False, True, False, ...]`). When passing
    `child_indices=[2,3,4]`, the return value of this method would be
    `parent_indices=[5,7,9]` (indexing starts at 0)`. Index 5 in the
    parent dataset corresponds to index 2 in the child dataset.

    Parameters
    ----------
    child: RTDC_Hierarchy
        RTDCBase hierarchy child to map from
    child_indices: 1d int ndarray
        integer indices in `child`

    Returns
    -------
    parent_indices: 1d int ndarray
        integer indices in `child.hparent`
    """
    parent = child.hparent
    # filters
    pf = parent.filter.all
    # indices corresponding to all child events
    idx = np.where(pf)[0]  # True means present in the child
    # indices corresponding to selected child events
    parent_indices = idx[child_indices]
    return parent_indices


def map_indices_child2root(child, child_indices):
    """Map RTDC_Hierarchy event indices to root RTDCBase

    Like :func:`map_indices_child2parent`, but map the
    child indices to the root parent.

    Parameters
    ----------
    child: RTDC_Hierarchy
        RTDCBase hierarchy child to map from
    child_indices: 1d ndarray
        integer indices in `child`

    Returns
    -------
    root_indices: 1d ndarray
        integer indices in the child's root parent
        (not necessarily the indices of `child.hparent`)
    """
    while True:
        indices = map_indices_child2parent(child=child,
                                           child_indices=child_indices)
        if child.hparent.format == "hierarchy":
            child = child.hparent
            child_indices = indices
        else:
            break
    return indices


def map_indices_parent2child(child, parent_indices):
    """Map parent RTDCBase event indices to RTDC_Hierarchy child

    Given a hierarchy child and indices defined for its `child.hparent`,
    return the corresponding indices for the `child`.

    Parameters
    ----------
    child: RTDC_Hierarchy
        RTDCBase hierarchy child to map to
    parent_indices: 1d ndarray
        integer indices in `child.hparent`

    Returns
    -------
    child_indices: 1d ndarray
        integer indices in `child`, corresponding to `parent_indices`
        in `child.hparent`
    """
    parent = child.hparent
    # this boolean array defines `child` in the parent
    pf = parent.filter.all
    # all event indices in parent that define `child`
    pf_loc = np.where(pf)[0]
    # boolean array with size `len(child)` indicating where the
    # `parent_indices` are set.
    same = np.isin(pf_loc, parent_indices)
    return np.where(same)[0]


def map_indices_root2child(child, root_indices):
    """Map root RTDCBase event indices to RTDC_Hierarchy child

    Like :func:`map_indices_parent2child`, but accepts the
    `root_indices` and map them to `child`.

    Parameters
    ----------
    child: RTDCBase
        RTDCBase hierarchy child to map to
    root_indices: 1d ndarray
        integer indices in the root parent of `child`

    Returns
    -------
    child_indices: 1d ndarray
        integer indices in `child`, corresponding to `root_indices`
        in `child`s root parent
    """
    # construct hierarchy tree containing only RTDC_Hierarchy instances
    hierarchy = [child]
    while True:
        if child.hparent.format == "hierarchy":
            # the parent is a hierarchy tree
            hierarchy.append(child.hparent)
            child = child.hparent
        else:
            break

    indices = root_indices
    for hp in hierarchy[::-1]:  # reverse order
        # For each hierarchy parent, map the indices down the
        # hierarchy tree.
        indices = map_indices_parent2child(child=hp,
                                           parent_indices=indices)
    return indices



"""
With basins, you can create analysis pipelines that result in output files
which, when opened in dclab, can access features stored in the input file
(without having to write those features to the output file).
"""
from __future__ import annotations

import abc
import numbers
import threading
from typing import Dict, List, Literal
import uuid
import warnings
import weakref

import numpy as np

from ..util import copy_if_needed


class BasinFeatureMissingWarning(UserWarning):
    """Used when a badin feature is defined but not stored"""


class CyclicBasinDependencyFoundWarning(UserWarning):
    """Used when a basin is defined in one of its sub-basins"""


class BasinmapFeatureMissingError(KeyError):
    """Used when one of the `basinmap` features is not defined"""
    pass


class BasinNotAvailableError(BaseException):
    """Used to identify situations where the basin data is not available"""
    pass


class BasinAvailabilityChecker(threading.Thread):
    """Helper thread for checking basin availability in the background"""
    def __init__(self, basin, *args, **kwargs):
        super(BasinAvailabilityChecker, self).__init__(*args, daemon=True,
                                                       **kwargs)
        self.basin = basin

    def run(self):
        self.basin.is_available()


class Basin(abc.ABC):
    """A basin represents data from an external source

    The external data must be a valid RT-DC dataset, subclasses
    should ensure that the corresponding API is available.
    """
    id_getters = {}

    def __init__(self,
                 location: str,
                 name: str = None,
                 description: str = None,
                 features: List[str] = None,
                 measurement_identifier: str = None,
                 mapping: Literal["same",
                                  "basinmap0",
                                  "basinmap1",
                                  "basinmap2",
                                  "basinmap3",
                                  "basinmap4",
                                  "basinmap5",
                                  "basinmap6",
                                  "basinmap7",
                                  "basinmap8",
                                  "basinmap9",
                                  ] = "same",
                 mapping_referrer: Dict = None,
                 ignored_basins: List[str] = None,
                 key: str = None,
                 **kwargs):
        """

        Parameters
        ----------
        location: str
            Location of the basin, this can be a path or a URL, depending
            on the implementation of the subclass
        name: str
            Human-readable name of the basin
        description: str
            Lengthy description of the basin
        features: list of str
            List of features this basin provides; This list is enforced,
            even if the basin actually contains more features.
        measurement_identifier: str
            A measurement identifier against which to check the basin.
            If this is set to None (default), there is no certainty
            that the downstream dataset is from the same measurement.
        mapping: str
            Which type of mapping to use. This can be either "same"
            when the event list of the basin is identical to that
            of the dataset defining the basin, or one of the "basinmap"
            features (e.g. "basinmap1") in cases where the dataset consists
            of a subset of the events of the basin dataset. In the latter
            case, the feature defined by `mapping` must be present in the
            dataset and consist of integer-valued indices (starting at 0)
            for the basin dataset.
        mapping_referrer: dict-like
            Dict-like object from which "basinmap" features can be obtained
            in situations where `mapping != "same"`. This can be a simple
            dictionary of numpy arrays or e.g. an instance of
            :class:`.RTDCBase`.
        ignored_basins: list of str
            List of basins to ignore in subsequent basin instantiations
        key: str
            Unique key to identify this basin; normally computed from
            a JSON dump of the basin definition. A random string is used
            if None is specified.
        kwargs:
            Additional keyword arguments passed to the `load_dataset`
            method of the `Basin` subclass.

        .. versionchanged: 0.58.0

            Added the `mapping` keyword argument to support basins
            with a superset of events.
        """
        #: location of the basin (e.g. path or URL)
        self.location = location
        #: user-defined name of the basin
        self.name = name
        #: lengthy description of the basin
        self.description = description
        # defining key of the basin
        self.key = key or str(uuid.uuid4())
        # features this basin provides
        self._features = features
        #: measurement identifier of the referencing dataset
        self.measurement_identifier = measurement_identifier
        self._measurement_identifier_verified = False
        #: ignored basins
        self.ignored_basins = ignored_basins or []
        #: additional keyword arguments passed to the basin
        self.kwargs = kwargs
        #: Event mapping strategy. If this is "same", it means that the
        #: referring dataset and the basin dataset have identical event
        #: indices. If `mapping` is e.g. `basinmap1` then the mapping of the
        #: indices from the basin to the referring dataset is defined in
        #: `self.basinmap` (copied during initialization of this class from
        #: the array in the key `basinmap1` from the dict-like object
        #: `mapping_referrer`).
        self.mapping = mapping or "same"
        self._basinmap = None  # see `basinmap` property
        # Create a weakref to the original referrer: If it is an instance
        # of RTDCBase, then garbage collection can clean up properly and
        # the basin instance has no reason to exist without the referrer.
        if self.mapping != "same":
            self._basinmap_referrer = weakref.ref(mapping_referrer)
        else:
            self._basinmap_referrer = None
        self._ds = None
        # perform availability check in separate thread
        self._av_check_lock = threading.Lock()
        self._av_check = BasinAvailabilityChecker(self)
        self._av_check.start()

    def __repr__(self):
        options = [
            self.name,
            f"mapped {self.mapping}" if self.mapping != "same" else "",
            f"features {self._features}" if self.features else "full-featured",
            f"location {self.location}",
        ]
        opt_str = ", ".join([o for o in options if o])

        return f"<{self.__class__.__name__} ({opt_str}) at {hex(id(self))}>"

    def _assert_measurement_identifier(self):
        """Make sure the basin matches the measurement identifier
        """
        if not self.verify_basin(run_identifier=True):
            raise KeyError(f"Measurement identifier of basin {self.ds} "
                           f"({self.get_measurement_identifier()}) does "
                           f"not match {self.measurement_identifier}!")

    @property
    def basinmap(self):
        """Contains the indexing array in case of a mapped basin"""
        if self._basinmap is None:
            if self.mapping != "same":
                try:
                    basinmap = self._basinmap_referrer()[self.mapping]
                except (KeyError, RecursionError):
                    raise BasinmapFeatureMissingError(
                        f"Could not find the feature '{self.mapping}' in the "
                        f"dataset or any of its basins. This suggests that "
                        f"this feature was never saved anywhere. Please check "
                        f"the input files.")
                #: `basinmap` is an integer array that maps the events from the
                #: basin to the events of the referring dataset.
                self._basinmap = np.array(basinmap,
                                          dtype=np.uint64,
                                          copy=True)
            else:
                self._basinmap = None
        return self._basinmap

    @property
    @abc.abstractmethod
    def basin_format(self):
        """Basin format (:class:`.RTDCBase` subclass), e.g. "hdf5" or "s3"
        """
        # to be implemented in subclasses

    @property
    @abc.abstractmethod
    def basin_type(self):
        """Storage type to use (e.g. "file" or "remote")"""
        # to be implemented in subclasses

    @property
    def ds(self):
        """The :class:`.RTDCBase` instance represented by the basin"""
        if self._ds is None:
            if not self.is_available():
                raise BasinNotAvailableError(f"Basin {self} is not available!")
            self._ds = self.load_dataset(self.location, **self.kwargs)
            self._ds.ignore_basins(self.ignored_basins)
        return self._ds

    @property
    def features(self):
        """Features made available by the basin

        .. versionchanged: 0.56.0

           Return nested basin features
        """
        if self._features is None:
            if self.is_available():
                # If features are not specified already, either by previous
                # call to this method or during initialization from basin
                # definition, then make the innate and *all* the basin
                # features available.
                self._features = sorted(set(self.ds.features_innate
                                            + self.ds.features_basin))
            else:
                self._features = []
        return self._features

    def as_dict(self):
        """Return basin kwargs for :func:`RTDCWriter.store_basin`

        Note that each subclass of :class:`.RTDCBase` has its own
        implementation of :func:`.RTDCBase.basins_get_dicts` which
        returns a list of basin dictionaries that are used to
        instantiate the basins in :func:`RTDCBase.basins_enable`.
        This method here is only intended for usage with
        :func:`RTDCWriter.store_basin`.
        """
        return {
            "basin_name": self.name,
            "basin_type": self.basin_type,
            "basin_format": self.basin_format,
            "basin_locs": [self.location],
            "basin_descr": self.description,
            "basin_feats": self.features,
            "basin_map": self.basinmap,
        }

    def close(self):
        """Close any open file handles or connections"""
        if self._ds is not None:
            self._ds.close()
        self._av_check.join(0.5)

    def get_feature_data(self, feat):
        """Return an object representing feature data of the basin"""
        self._assert_measurement_identifier()
        return self.ds[feat]

    def get_measurement_identifier(self):
        """Return the identifier of the basin dataset"""
        return self.ds.get_measurement_identifier()

    @abc.abstractmethod
    def is_available(self):
        """Return True if the basin is available"""

    @abc.abstractmethod
    def _load_dataset(self, location, **kwargs):
        """Subclasses should return an instance of :class:`.RTDCBase`"""

    def load_dataset(self, location, **kwargs):
        """Return an instance of :class:`.RTDCBase` for this basin

        If the basin mapping (`self.mapping`) is not the same as
        the referencing dataset
        """
        ds = self._load_dataset(location, **kwargs)
        if self.mapping != "same":
            # The array `self.basinmap` may contain duplicate elements,
            # which is why we cannot use hierarchy children to access the
            # data (sometimes the data must be blown-up rather than gated).
            ds_bn = BasinProxy(ds=ds, basinmap=self.basinmap)
        else:
            ds_bn = ds
        return ds_bn

    def verify_basin(self, run_identifier=True, availability=True):
        if not availability:
            warnings.warn("The keyword argument 'availability' is "
                          "deprecated, because it can lead to long waiting "
                          "times with many unavailable basins.",
                          DeprecationWarning)
        if availability:
            check_avail = self.is_available()
        else:
            check_avail = True

        # Only check for run identifier if requested and if the availability
        # check did not fail.
        if run_identifier and check_avail:
            if not self._measurement_identifier_verified:
                if self.measurement_identifier is None:
                    # No measurement identifier was presented by the
                    # referencing dataset. Don't perform any checks.
                    self._measurement_identifier_verified = True
                else:
                    if self.mapping == "same":
                        # When we have identical mapping, then the measurement
                        # identifier has to match exactly.
                        verifier = str.__eq__
                    else:
                        # When we have non-identical mapping (e.g. exported
                        # data), then the measurement identifier has to
                        # partially match.
                        verifier = str.startswith
                    self._measurement_identifier_verified = verifier(
                        self.measurement_identifier,
                        self.get_measurement_identifier()
                    )
            check_rid = self._measurement_identifier_verified
        else:
            check_rid = True

        return check_rid and check_avail


class BasinProxy:
    def __init__(self, ds, basinmap):
        """Proxy for accessing data in basin datasets

        The idea of a basin proxy is to give access to the data of an
        :class:`.RTDCBase` that is mapped, i.e. the indices defined for
        the basin do not coincide with the indices in the downstream
        dataset.

        This class achieves two things:
        1. Subset indexing: For every event in the downstream dataset, there
           is *only* one corresponding event in the basin dataset. This
           could also be achieved via hierarchy children
           (:class:`RTDCHierarchy`).
        2. Blown indexing: Two different events in the downstream dataset
           can refer to one event in the basin dataset. I.e. the basin
           dataset contains fewer events than the downstream dataset,
           because e.g. it is a raw image recording series that has been
           processed and multiple events were found in one frame.

        Parameters
        ----------
        ds: RTDCBase
            the basin dataset
        basinmap: np.ndarray
            1D integer indexing array that maps the events of the basin
            dataset to the downstream dataset
        """
        self.ds = ds
        self.basins_get_dicts = ds.basins_get_dicts
        self.basinmap = basinmap
        self._features = {}

    def __contains__(self, item):
        return item in self.ds

    def __getattr__(self, item):
        if item in [
            "basins",
            "close",
            "features",
            "features_ancillary",
            "features_basin",
            "features_innate",
            "features_loaded",
            "features_local",
            "features_scalar",
            "get_measurement_identifier",
            "ignore_basins",
        ]:
            return getattr(self.ds, item)
        else:
            raise AttributeError(
                f"BasinProxy does not implement {item}")

    def __getitem__(self, feat):
        if feat not in self._features:
            feat_obj = BasinProxyFeature(feat_obj=self.ds[feat],
                                         basinmap=self.basinmap)
            self._features[feat] = feat_obj
        return self._features[feat]

    def __len__(self):
        return len(self.basinmap)


class BasinProxyFeature(np.lib.mixins.NDArrayOperatorsMixin):
    def __init__(self, feat_obj, basinmap):
        """Wrap around a feature object, mapping it upon data access"""
        self.feat_obj = feat_obj
        self.basinmap = basinmap
        self._cache = None
        self.is_scalar = bool(len(self.feat_obj.shape) == 1)

    def __array__(self, dtype=None, copy=copy_if_needed, *args, **kwargs):
        if self._cache is None and self.is_scalar:
            self._cache = self.feat_obj[:][self.basinmap]
        else:
            # This is dangerous territory in terms of memory usage
            out_arr = np.empty((len(self.basinmap),) + self.feat_obj.shape[1:],
                               dtype=dtype or self.feat_obj.dtype,
                               *args, **kwargs)
            for ii, idx in enumerate(self.basinmap):
                out_arr[ii] = self.feat_obj[idx]
            return out_arr
        return np.array(self._cache, copy=copy)

    def __getattr__(self, item):
        if item in [
            "dtype",
            "shape",
            "size",
        ]:
            return getattr(self.feat_obj, item)
        else:
            raise AttributeError(
                f"BasinProxyFeature does not implement {item}")

    def __getitem__(self, index):
        if self._cache is None and isinstance(index, numbers.Integral):
            # single index, cheap operation
            return self.feat_obj[self.basinmap[index]]
        elif not self.is_scalar:
            # image, mask, etc
            if isinstance(index, slice) and index == slice(None):
                indices = self.basinmap
            else:
                indices = self.basinmap[index]
            out_arr = np.empty((len(indices),) + self.feat_obj.shape[1:],
                               dtype=self.feat_obj.dtype)
            for ii, idx in enumerate(indices):
                out_arr[ii] = self.feat_obj[idx]
            return out_arr
        else:
            # sets the cache if not already set
            return self.__array__()[index]

    def __len__(self):
        return len(self.basinmap)


def basin_priority_sorted_key(bdict: Dict):
    """Yield a sorting value for a given basin that can be used with `sorted`

    Basins are normally stored in random order in a dataset. This method
    brings them into correct order, prioritizing:

    - type: "file" over "remote"
    - format: "HTTP" over "S3" over "dcor"
    - mapping: "same" over anything else
    """
    srt_type = {
        "internal": "a",
        "file": "b",
        "remote": "c",
    }.get(bdict.get("type"), "z")

    srt_format = {
        "h5dataset": "a",
        "hdf5": "b",
        "http": "c",
        "s3": "d",
        "dcor": "e",
    }.get(bdict.get("format"), "z")

    mapping = bdict.get("mapping", "same")  # old dicts don't have "mapping"
    srt_map = "a" if mapping == "same" else mapping

    return srt_type + srt_format + srt_map


class InternalH5DatasetBasin(Basin):
    basin_format = "h5dataset"
    basin_type = "internal"

    def __init__(self, *args, **kwargs):
        super(InternalH5DatasetBasin, self).__init__(*args, **kwargs)
        if self.mapping == "same":
            raise ValueError(
                "'internal' basins must be instantiated with `mapping`. "
                "If you are not doing that, then you probably don't need "
                "them.")
        if self._features is None:
            raise ValueError("You must specify features when defining "
                             "internal basins.")
        # Redefine the features if necessary
        h5root = self._basinmap_referrer().h5file
        available_features = []
        for feat in self._features:
            if self.location in h5root and feat in h5root[self.location]:
                available_features.append(feat)
            else:
                warnings.warn(
                    f"Feature '{feat}' is defined as an internal basin, "
                    f"but it cannot be found in '{self.location}'.",
                    BasinFeatureMissingWarning)
        self._features.clear()
        self._features += available_features

    def _load_dataset(self, location, **kwargs):
        from .fmt_dict import RTDC_Dict
        # get the h5file object
        h5root = self._basinmap_referrer().h5file
        ds_dict = {}
        for feat in self.features:
            ds_dict[feat] = h5root[self.location][feat]
        return RTDC_Dict(ds_dict)

    def is_available(self):
        return bool(self._features)

    def verify_basin(self, *args, **kwargs):
        """It's not necessary to verify internal basins"""
        return True


def get_basin_classes():
    bc = {}
    for b_cls in Basin.__subclasses__():
        if hasattr(b_cls, "basin_format"):
            bc[b_cls.basin_format] = b_cls
    return bc



"""Generate in-silico .rtdc files from FEM simulation data

The simulation dataset must be an HDF5 file with a specific structure
that contains the results of FEM simulations. The original HDF5 files
were provided by Lucas Wittwer.

This script generates multiple (default: 5) .rtdc datasets using the
shape data from the FEM simulations. The datasets are generated from
all simulations, but each with a random pixel offset when casting the
contour data onto a regular grid. In addition, an "original" .rtdc
dataset is generated, which consists of the original simulation
parameters.

Supported simulation modalities:
- 2D axis-symmetric (used for the first emodulus LUT in dclab)

Testing: `pytest fem2rtdc.py`
"""
import argparse
import copy
import pathlib

import cv2
import h5py
import numpy as np

import dclab
from dclab.features.contour import get_contour
from dclab.features.volume import get_volume


def features_from_2daxis_simulation(fem_cont_2daxis, pixel_size=0.34,
                                    shape=(80, 250), seed=42):
    """Compute dclab features from the FEM contour

    Map the FEM contour onto a regular grid and displace it
    randomly by up to +/- 0.5 pixels (used for investigating
    pixelation effects).

    Parameters
    ----------
    fem_cont_2daxis: ndarray of shape (2, N)
        The 2D axis-symmetric FEM contour. It should only cover
        180 degrees of the entire contour.
    pixel_size: float
        Detector pixel size [um]
    shape: tuple of int
        Detector shape / ROI in pixels
    seed: int or None
        Seed for numpy.random which is used for the random
        displacements. You may set this to `None` and set the
        seed yourself.
    """
    # Subtract the mean in x direction; this is important for mapping
    # out the mask from the contour (too large values are outside the image)
    fem_cont_2daxis[0] -= np.mean(fem_cont_2daxis[0])
    assert np.all(fem_cont_2daxis[1] >= 0), "2D contour data must be positive"
    if seed is not None:
        np.random.seed(seed)
    offx = np.random.uniform(0, 1)
    offy = np.random.uniform(0, 1)
    # make contour circular
    simsym = fem_cont_2daxis
    sh = simsym.shape[1]
    cont_sim = np.zeros((simsym.shape[1] * 2 - 2, 2), dtype=np.float32)
    cont_sim[:sh, 1] = simsym[0]
    cont_sim[:sh, 0] = simsym[1]
    cont_sim[sh:, 1] = simsym[0, 1:-1][::-1]
    cont_sim[sh:, 0] = -simsym[1, 1:-1][::-1]
    # convert um to pixels
    cont_sim /= pixel_size
    # center around center of channel
    cont_sim[:, 0] += shape[0] // 2 - np.mean(cont_sim[:, 0]) / 2 + offx
    cont_sim[:, 1] += shape[1] // 2 - np.mean(cont_sim[:, 1]) / 2 + offy
    # put on grid
    mask = np.zeros(shape)
    xmin = int(np.floor(np.min(cont_sim[:, 0])))
    xmax = int(np.ceil(np.max(cont_sim[:, 0])))
    ymin = int(np.floor(np.min(cont_sim[:, 1])))
    ymax = int(np.ceil(np.max(cont_sim[:, 1])))

    xarr = np.arange(xmin, xmax)
    yarr = np.arange(ymin, ymax)
    xm, ym = np.meshgrid(xarr, yarr, indexing="ij")
    points = np.zeros((xm.size, 2), dtype=int)
    points[:, 0] = xm.flatten()
    points[:, 1] = ym.flatten()
    pip = dclab.polygon_filter.points_in_poly(points, cont_sim)
    mask[points[:, 0], points[:, 1]] = pip

    return features_from_mask(mask, pixel_size=pixel_size)


def features_from_mask(mask, pixel_size=0.34):
    """Compute dclab features from a binary mask image

    This essentially mimicks what Shape-In does with OpenCV.

    Parameters
    ----------
    mask: 2d boolean ndarray
        binary mask image
    pixel_size: float
        Detector pixel size [um]
    """
    cont_raw = get_contour(mask)
    # this is how Shape-In does things
    cont = cv2.convexHull(cont_raw)
    arc = cv2.arcLength(cont, True)
    mu = cv2.moments(cont, False)
    pos_x = (mu["m10"] / mu["m00"]) * pixel_size
    pos_y = (mu["m01"] / mu["m00"]) * pixel_size
    feats = {
        # scalar features
        "area_um": mu["m00"] * pixel_size ** 2,
        "deform": 1 - 2.0 * np.sqrt(np.pi * mu["m00"]) / arc,
        "inert_ratio_cvx": np.sqrt(mu["mu20"] / mu["mu02"]),
        "pos_x": pos_x,
        "pos_y": pos_y,
        "volume": get_volume(cont_raw, pos_x, pos_y, pixel_size),
        # image data
        "image": np.array(254 * mask, dtype=np.uint8),
        "mask": mask,
    }
    return feats


def generate_rtdc_from_simulation(path, repetitions=5, pixel_size=0.34):
    """Generate in-silico .rtdc files from a FEM simulation file

    In addition to the in-silico datasets, an "original" .rtdc
    dataset is generated, which consists of the original simulation
    parameters.

    Parameters
    ----------
    path: pathlib.Path or str
        The simulation dataset must be an HDF5 file with a specific
        structure that contains the results of FEM simulations. The
        original HDF5 files were provided by Lucas Wittwer.
    repetitions: int
        Number of in-silico .rtdc files to generate. The datasets are
        generated from all simulations, but each with a random pixel
        offset when casting the contour data onto a regular grid.
    pixel_size: float
        Detector pixel size [um]
    """
    path = pathlib.Path(path)

    writers_mapped = []
    for ii in range(repetitions):
        pi = path.with_suffix(".map{}.rtdc".format(ii + 1))
        if pi.exists():
            pi.unlink()
        writers_mapped.append(dclab.RTDCWriter(pi))

    # Exact data from simulations
    path_orig = path.with_suffix(".orig.rtdc")
    if path_orig.exists():
        path_orig.unlink()
    writer_orig = dclab.RTDCWriter(path_orig)

    # Extract dataset from simulation contour
    with h5py.File(path, "r") as h5:
        # determine total size
        ntot = 0
        for dk in h5.keys():
            for _ in h5[dk]:
                ntot += 1

        dkeys = sorted(h5.keys(), key=lambda x: int(x))

        # metadata
        meta = {
            "setup": {
                "channel width": h5.attrs["channel_width"],
                "flow rate": h5.attrs["flow_rate"],
                "chip region": "channel",
                "medium": "other",
                "software version": "dclab {}".format(dclab.__version__)
            },
            "imaging": {
                "pixel size": pixel_size,
                "roi size x": 250,
                "roi size y": 80,
            },
            "experiment": {
                "date": h5.attrs["date"],
                "time": "00:00:00",
                "event count": ntot,
            },
        }

        print("Converting original simulation data to .rtdc.")
        # Write unperturbed original data
        mo = copy.deepcopy(meta)
        mo["experiment"]["run index"] = 1
        mo["experiment"]["sample"] = "Simulation reference"
        writer_orig.store_metadata(mo)

        for dk in dkeys:
            for sk in h5[dk]:
                sim = h5[dk][sk]
                feats_orig = {
                    "area_um": sim.attrs["area"],
                    "deform": sim.attrs["deformation"],
                    "volume": sim.attrs["volume"],
                    "emodulus": sim.parent.attrs["emodulus"] / 1000,
                }
                for feat in feats_orig:
                    writer_orig.store_feature(feat, feats_orig[feat])

        # Write randomly shifted data
        for ii, wi in enumerate(writers_mapped):
            mi = copy.deepcopy(meta)
            mi["experiment"]["run index"] = ii + 1
            mi["experiment"]["sample"] = "Contour mapping {}".format(ii + 1)
            wi.store_metadata(mi)

        ss = 0
        for jj, wi in enumerate(writers_mapped):
            np.random.seed(jj)
            for dk in dkeys:
                for sk in h5[dk]:
                    ss += 1
                    print(
                        f"Creating mask-shifted data {jj+1}: "
                        + f"{ss / ntot / repetitions * 100:.1f}%", end="\r")
                    sim = h5[dk][sk]
                    simsym = sim["coords"][:]
                    feats = features_from_2daxis_simulation(
                        simsym, pixel_size=pixel_size, seed=None)
                    for feat in feats:
                        wi.store_feature(feat, feats[feat])

        for wr in writers_mapped + [writer_orig]:
            wr.rectify_metadata()
            wr.version_brand()
            wr.h5file.close()


def test_features_from_2daxis_simulation():
    feats = features_from_2daxis_simulation(
        fem_cont_2daxis=np.copy(test_coords),
        pixel_size=0.34,
        shape=(80, 250),
        seed=42)
    assert np.allclose(feats['area_um'], 33.69740000000001)
    assert np.allclose(feats['deform'], 0.1168877846976153)
    assert np.allclose(feats['inert_ratio_cvx'], 1.285950706011422)
    assert np.allclose(feats['pos_x'], 43.4609033733562)
    assert np.allclose(feats['pos_y'], 13.717804459691251)
    assert np.allclose(feats['volume'], 125.581725903289)


def test_generate_rtdc_from_simulation(tmp_path):
    # create a test file
    path = tmp_path / "input_file.hdf5"
    with h5py.File(path, "w") as h5:
        h5["0/0/coords"] = test_coords
        h5.attrs.update({
            'authors': 'Someone',
            'channel_width': 20.0,
            'channel_width_unit': 'um',
            'commit_hash': '1233454',
            'date': '2019-11-13',
            'dimensionality': '2Daxis',
            'flow_rate': 0.04,
            'flow_rate_unit': 'uL/s',
            'fluid_density': 1000.0,
            'fluid_density_unit': 'kg m^-3',
            'fluid_viscosity': 0.015,
            'fluid_viscosity_unit': 'Pa s',
            'identifier': 'LE-2D-FEM-19',
            'method': 'FEM',
            'model': 'linear elastic',
            'publication': 'https://doi.org/10.1021/acsbiomaterials.6b00558',
            'software': 'custom C++ code',
            'solid_density': 1000.0,
            'solid_density_unit': 'kg m^-3',
            'summary': 'Test dataset'})
        h5["0"].attrs["emodulus"] = 740.83203
        h5["0"].attrs["emodulus_unit"] = "Pa"
        h5["0/0"].attrs["area"] = 36.42331
        h5["0/0"].attrs["area_unit"] = "um^2"
        h5["0/0"].attrs["deformation"] = 0.108169
        h5["0/0"].attrs["deformation_unit"] = ""
        h5["0/0"].attrs["radius"] = 3.282
        h5["0/0"].attrs["radius_unit"] = "um"
        h5["0/0"].attrs["sphere_volume"] = 148.0827
        h5["0/0"].attrs["sphere_volume_unit"] = "um^3"
        h5["0/0"].attrs["volume"] = 147.95604
        h5["0/0"].attrs["volume_unit"] = "um^3"
        h5["0/0/coords"].attrs["FIELD_0"] = "x"
        h5["0/0/coords"].attrs["FIELD_1"] = "y"
        h5["0/0/coords"].attrs["coords_unit"] = "um"

    feats1 = features_from_2daxis_simulation(
        fem_cont_2daxis=np.copy(test_coords),
        pixel_size=0.34,
        shape=(80, 250),
        seed=0)

    generate_rtdc_from_simulation(path, repetitions=5, pixel_size=0.34)
    with h5py.File(path.with_suffix(".map1.rtdc")) as h51:
        assert np.allclose(feats1['area_um'], h51["events/area_um"][0])
        assert np.allclose(feats1['deform'], h51["events/deform"][0])
        assert np.allclose(feats1['inert_ratio_cvx'],
                           h51["events/inert_ratio_cvx"][0])
        assert np.allclose(feats1['pos_x'], h51["events/pos_x"][0])
        assert np.allclose(feats1['pos_y'], h51["events/pos_y"][0])
        assert np.allclose(feats1['volume'], h51["events/volume"][0])


#: The first contour from https://doi.org/10.6084/m9.figshare.12155064.v4
test_coords = np.array(
    [[-2.647751, -2.647751, -2.647751, -2.647141, -2.646541,
      -2.645251, -2.643971, -2.642191, -2.640421, -2.638001,
      -2.635591, -2.632401, -2.629221, -2.625081, -2.620941,
      -2.6156511, -2.610371, -2.6037111, -2.597061, -2.588711,
      -2.5803611, -2.5700011, -2.5596411, -2.5470111, -2.5343711,
      -2.5190911, -2.5038111, -2.485591, -2.467371, -2.445271,
      -2.423161, -2.397831, -2.3724911, -2.343981, -2.315471,
      -2.284131, -2.252801, -2.216171, -2.179531, -2.134901,
      -2.090281, -2.032001, -1.973711, -1.900521, -1.827321,
      -1.744361, -1.661401, -1.568581, -1.4757711, -1.3715811,
      -1.267401, -1.152081, -1.036771, -0.90868104, -0.780591,
      -0.639801, -0.49901104, -0.34514102, -0.19128104, -0.02494103,
      0.14138897, 0.31868896, 0.49598897, 0.68322897, 0.870469,
      1.064479, 1.258499, 1.4578589, 1.6572189, 1.8587589,
      2.060299, 2.262629, 2.464949, 2.665009, 2.865059,
      3.0620189, 3.258979, 3.449929, 3.640879, 3.824039,
      4.007209, 4.180559, 4.353919, 4.513099, 4.672279,
      4.815419, 4.958549, 5.078679, 5.198809, 5.294079,
      5.389349, 5.455509, 5.521669, 5.5633388, 5.605009,
      5.622299, 5.639589],
     [0., 0.09835, 0.19869, 0.29785, 0.39701,
      0.49532, 0.59363, 0.69125, 0.78887, 0.88545,
      0.98203, 1.07731, 1.17258, 1.2662, 1.35981,
      1.4515, 1.54319, 1.63265, 1.7221, 1.809,
      1.89591, 1.97986, 2.06381, 2.14455, 2.2253,
      2.30242, 2.37953, 2.45276, 2.52598, 2.59443,
      2.66287, 2.72659, 2.79032, 2.84945, 2.90857,
      2.96377, 3.01897, 3.0693, 3.11963, 3.16366,
      3.2077, 3.24223, 3.27676, 3.29975, 3.32275,
      3.33584, 3.34893, 3.35274, 3.35654, 3.35024,
      3.34393, 3.32765, 3.31137, 3.28401, 3.25665,
      3.21855, 3.18045, 3.13136, 3.08227, 3.02287,
      2.96347, 2.89481, 2.82615, 2.74929, 2.67243,
      2.58959, 2.50675, 2.41919, 2.33163, 2.24136,
      2.15109, 2.05934, 1.96759, 1.87593, 1.78428,
      1.69324, 1.60219, 1.51283, 1.42347, 1.33594,
      1.24841, 1.16308, 1.07775, 0.99495, 0.91216,
      0.83188, 0.7516, 0.6742, 0.59681, 0.52143,
      0.44606, 0.37198, 0.29791, 0.22335, 0.1488,
      0.07659, 0.]])


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('input', type=str, help='Input path (.hdf5 file)')
    parser.add_argument('--repetitions', type=int, default=5,
                        help='Number of repetitions (grid mapping)')
    parser.add_argument('--pixelsize', type=float, default=0.34,
                        help='Detector pixel size [um]')

    args = parser.parse_args()

    generate_rtdc_from_simulation(path=pathlib.Path(args.input),
                                  repetitions=args.repetitions,
                                  pixel_size=args.pixelsize)



import pathlib
import socket
import tempfile
import zipfile

import numpy as np

from dclab.rtdc_dataset import fmt_tdms
from dclab import definitions as dfn
from dclab.rtdc_dataset.fmt_dcor import REQUESTS_AVAILABLE  # noqa: F401
from dclab.rtdc_dataset.fmt_s3 import BOTO3_AVAILABLE  # noqa: F401


with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
    try:
        # dcor.mpl.mpg.de
        s.connect(("130.183.206.45", 443))
    except (socket.gaierror, OSError):
        DCOR_AVAILABLE = False
    else:
        DCOR_AVAILABLE = True


with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
    try:
        # objectstore.hpccloud.mpcdf.mpg.de
        s.connect(("130.183.12.74", 443))
    except (socket.gaierror, OSError):
        DCOR_AVAILABLE = False


def calltracker(func):
    """Decorator to track how many times a function is called"""
    def wrapped(*args, **kwargs):
        wrapped.calls += 1
        return func(*args, **kwargs)
    wrapped.calls = 0
    return wrapped


def example_data_dict(size=100, keys=["area_um", "deform"]):
    """Example dict with which an RTDCBase can be instantiated.
    """
    ddict = {}
    for ii, key in enumerate(keys):
        if key in ["time", "frame"]:
            val = np.arange(size)
        elif key == "contour":
            cdata = []
            for ss in range(size):
                cont = np.array([5, 5,
                                 5, 6,
                                 5, 7,
                                 6, 7,
                                 7, 7,
                                 7, 6,
                                 7, 5,
                                 6, 5,
                                 5, 5]).reshape(-1, 2)
                cdata.append(cont + ss)
            val = cdata
        elif key in ["image", "image_bg"]:
            imdat = []
            for ss in range(size):
                data = np.arange(10 * 20).reshape(10, 20) + ss
                imdat.append(np.array(data, dtype=np.uint8))
            val = imdat
        elif key == "index":
            # index starts at 1
            val = np.arange(1, size+1)
        elif key == "trace":
            trdata = {}
            kk = 1
            for tr in dfn.FLUOR_TRACES:
                trac = np.arange(100 * size, dtype=np.int16).reshape(size, -1)
                trdata[tr] = trac - kk
                kk += 1
            val = trdata
        else:
            state = np.random.RandomState(size + ii)
            val = state.random_sample(size)

        norm = False
        if key == "area_um":
            norm = 400
        elif key in ["deform", "circ"]:
            norm = .02
        if norm:
            vmin, vmax = val.min(), val.max()
            val = (val - vmin) / (vmax - vmin) * norm

        ddict[key] = val

    return ddict


def find_data(path):
    """Find tdms and rtdc data files in a directory"""
    path = pathlib.Path(path)
    tdmsfiles = fmt_tdms.get_tdms_files(path)
    rtdcfiles = [r for r in path.rglob("*.rtdc") if r.is_file()]
    files = [pathlib.Path(ff) for ff in rtdcfiles + tdmsfiles]
    return files


def retrieve_data(zip_file):
    """Extract contents of data zip file and return data files
    """
    zpath = pathlib.Path(__file__).resolve().parent / "data" / zip_file
    # unpack
    arc = zipfile.ZipFile(str(zpath))

    # extract all files to a temporary directory
    edest = tempfile.mkdtemp(prefix=zpath.name)
    arc.extractall(edest)

    # Load RT-DC dataset
    # find tdms files
    datafiles = find_data(edest)

    if len(datafiles) == 1:
        datafiles = datafiles[0]

    return datafiles



import numbers

import numpy as np


def f1dfloatduple(value):
    """Tuple of two floats (duple)"""
    if np.array(value).ndim != 1:
        raise ValueError(f"Value is not 1 dimensional, got {value}!")
    value = tuple(float(i) for i in value)
    if len(value) != 2:
        raise ValueError(f"Value must be of length two, "
                         f"got length {len(value)}!")
    return value


def f2dfloatarray(value):
    """numpy floating point array"""
    return np.array(value, dtype=np.float64)


def fbool(value):
    """boolean"""
    if isinstance(value, str):
        value = value.lower()
        if value == "false":
            value = False
        elif value == "true":
            value = True
        elif value:
            value = bool(float(value))
        else:
            raise ValueError("Empty string provided for fbool!")
    else:
        value = bool(float(value))
    return value


def fboolorfloat(value):
    """Bool or float"""
    if isinstance(value, (str, bool)) or value == 0:
        return fbool(value)
    elif isinstance(value, (int, float)):
        return float(value)
    else:
        raise ValueError(f"Value could not be converted to bool "
                         f"or float, got {value}!")


def fint(value):
    """integer"""
    if isinstance(value, str):
        # strings might have been saved wrongly as booleans
        value = value.lower()
        if value == "false":
            value = 0
        elif value == "true":
            value = 1
        elif value:
            value = int(float(value))
        else:
            raise ValueError("Empty string provided for fint!")
    else:
        value = int(float(value))
    return value


def fintlist(alist):
    """A list of integers"""
    outlist = []
    if not isinstance(alist, (list, tuple)):
        # we have a string (comma-separated integers)
        alist = alist.strip().strip("[] ").split(",")
    for it in alist:
        if it:
            outlist.append(fint(it))
    return outlist


def lcstr(astr):
    """lower-case string"""
    return astr.lower()


#: maps functions to their expected output types
func_types = {
    f1dfloatduple: (tuple, np.ndarray),
    f2dfloatarray: np.ndarray,
    fbool: (bool, np.bool_),
    fboolorfloat: (bool, np.bool_, float),
    fint: numbers.Integral,
    fintlist: list,
    float: numbers.Number,
    lcstr: str}



import copy

from .meta_parse import (
    fbool, fint, fintlist, func_types, lcstr, f1dfloatduple, fboolorfloat
)

#: All configuration keywords editable by the user
CFG_ANALYSIS = {
    # filtering parameters
    "filtering": [
        ["hierarchy parent", str, "Hierarchy parent of the dataset"],
        ["remove invalid events", fbool, "Remove events with inf/nan values"],
        ["enable filters", fbool, "Enable filtering"],
        ["limit events", fint, "Upper limit for number of filtered events"],
        ["polygon filters", fintlist, "Polygon filter indices"],
    ],
    # Addition user-defined data
    "calculation": [
        ["emodulus lut", str, "Look-up table identifier"],
        ["emodulus medium", str, "Medium used (e.g. '0.49% MC-PBS')"],
        ["emodulus temperature", float, "Chip temperature [°C]"],
        ["emodulus viscosity", float, "Viscosity [Pa*s] if 'medium' unknown"],
        ["emodulus viscosity model", str, "Viscosity model for known media"],
        ["crosstalk fl21", float, "Fluorescence crosstalk, channel 2 to 1"],
        ["crosstalk fl31", float, "Fluorescence crosstalk, channel 3 to 1"],
        ["crosstalk fl12", float, "Fluorescence crosstalk, channel 1 to 2"],
        ["crosstalk fl32", float, "Fluorescence crosstalk, channel 3 to 2"],
        ["crosstalk fl13", float, "Fluorescence crosstalk, channel 1 to 3"],
        ["crosstalk fl23", float, "Fluorescence crosstalk, channel 2 to 3"],
    ]
}

#: All read-only configuration keywords for a measurement
CFG_METADATA = {
    # All parameters related to the actual experiment
    "experiment": [
        ["date", str, "Date of measurement ('YYYY-MM-DD')"],
        ["event count", fint, "Number of recorded events"],
        ["run index", fint, "Index of measurement run"],
        ["run identifier", str, "Unique measurement identifier"],
        ["sample", str, "Measured sample or user-defined reference"],
        ["time", str, "Start time of measurement ('HH:MM:SS[.S]')"],
        ["timestamp", float, "Start of measurement in unix time [s]"],
    ],
    # All special keywords related to RT-FDC
    # This section should not be present for regular RT-DC measurements.
    "fluorescence": [
        # The baseline offset was introduced in 0.33.0. It is added to
        # the trace data to obtain the actual signal used for data
        # processing (e.g. obtaining the fl1_max feature). This is more
        # robust than adding the offset directly to the trace data, because
        # of the possibility of integer overflows. Furthermore, DCKit can
        # set this parameter without modifying the original trace data
        # to correct/remove negative trace data
        # (see https://github.com/DC-analysis/dclab/issues/101).
        # Note that traces accessed from RTDCBase instances are never
        # background-corrected!
        ["baseline 1 offset", fint, "Baseline offset channel 1"],
        ["baseline 2 offset", fint, "Baseline offset channel 2"],
        ["baseline 3 offset", fint, "Baseline offset channel 3"],
        ["bit depth", fint, "Trace bit depth"],
        # If a fluorescence channel is used, a channel name *must* be
        # present. If a channel is not used, the channel name *must not*
        # be present. E.g. if only channels 1 and 2 are used, but there
        # are three channels present, then `channel count` is two,
        # `channels installed` is three, and `channel 3 name` is not set.
        ["channel 1 name", str, "FL1 description"],
        ["channel 2 name", str, "FL2 description"],
        ["channel 3 name", str, "FL3 description"],
        ["channel count", fint, "Number of active channels"],
        ["channels installed", fint, "Number of available channels"],
        # In contrast to `channel ? name`, the laser power *may*
        # be present (but must be set to 0), if a laser line is not used.
        ["laser 1 lambda", float, "Laser 1 wavelength [nm]"],
        ["laser 1 power", float, "Laser 1 output power [%]"],
        ["laser 2 lambda", float, "Laser 2 wavelength [nm]"],
        ["laser 2 power", float, "Laser 2 output power [%]"],
        ["laser 3 lambda", float, "Laser 3 wavelength [nm]"],
        ["laser 3 power", float, "Laser 3 output power [%]"],
        ["laser count", fint, "Number of active lasers"],
        ["lasers installed", fint, "Number of available lasers"],
        ["sample rate", fint, "Trace sample rate [Hz]"],
        ["samples per event", fint, "Samples per event"],
        ["signal max", float, "Upper voltage detection limit [V]"],
        ["signal min", float, "Lower voltage detection limit [V]"],
        ["trace median", fint, "Rolling median filter size for traces"],
    ],
    # All tdms-related parameters
    "fmt_tdms": [
        ["video frame offset", fint, "Missing events at beginning of video"],
    ],
    # All imaging-related keywords
    "imaging": [
        ["flash device", str, "Light source device type"],  # e.g. green LED
        ["flash duration", float, "Light source flash duration [µs]"],
        ["frame rate", float, "Imaging frame rate [Hz]"],
        ["pixel size", float, "Pixel size [µm]"],
        ["roi position x", fint, "Image x coordinate on sensor [px]"],
        ["roi position y", fint, "Image y coordinate on sensor [px]"],
        ["roi size x", fint, "Image width [px]"],
        ["roi size y", fint, "Image height [px]"],
    ],
    # All parameters for online contour extraction from the event images
    "online_contour": [
        # The option "bg empty" was introduced in dclab 0.34.0 and
        # Shape-In 2.2.2.5.
        # Shape-In  writes to the "shapein-warning" log if there are
        # frames with event images (non-empty frames) that had to be
        # used for background correction.
        ["bg empty", fbool, "Background correction from empty frames only"],
        ["bin area min", fint, "Minium pixel area of binary image event"],
        ["bin kernel", fint, "Disk size for binary closing of mask image"],
        ["bin threshold", fint, "Threshold for mask from bg-corrected image"],
        ["image blur", fint, "Odd sigma for Gaussian blur (21x21 kernel)"],
        ["no absdiff", fbool, "Do not use OpenCV 'absdiff' for bg-correction"],
    ],
    # All online-filter-related keywords (box filters, soft limit, and
    # polygons are handled in `meta_logic`).
    # Note that "soft limit" means that the event is still included in
    # the dataset, but is not counted for "target event count". On the
    # other hand, "hard limit" means that only those events that are
    # within that filter are written to the dataset.
    "online_filter": [
        # "target*" is only set if measurement is stopped automatically.
        # "target*" is not necessarily reached (e.g. user aborted).
        ["target duration", float, "Target measurement duration [min]"],
        ["target event count", fint, "Target event count for online gating"],
    ],
    # Data processing pipeline
    "pipeline": [
        ["dcnum background", str, "Background ID"],
        ["dcnum data", str, "Data ID"],
        ["dcnum feature", str, "Feature extractor ID"],
        ["dcnum gate", str, "Gating ID"],
        ["dcnum generation", str, "Generation ID"],
        # The hash is computed by joining the other pipeline identifiers with
        # the "|" character in this order: generation, data, background,
        # segmenter, feature, gate.
        ["dcnum hash", str, "Hash"],
        ["dcnum mapping", str, "Event mapping from original dataset"],
        ["dcnum segmenter", str, "Segmenter ID"],
        # The event yield the number of events the pipeline produced and
        # helps identify files that were e.g. exported from original files.
        ["dcnum yield", fint, "Event yield"],
    ],
    # All qpi-related keywords
    "qpi": [
        # experiment-related qpi metadata, see qpretrieve for details
        ["wavelength", float, "Imaging wavelength [nm]"],
        ["medium index", float, "Refractive index of medium"],
        ["pixel size raw", float, "Hologram pixel size [µm]."],
        # post-analysis-related qpi metadata
        ["software version", str, "Software version(s)"],
        # How background image was created 'experimental' or
        # computation method e.g. 'sparsemed'
        ["bg method", str, "Background computation method"],
        # calculation of pha and amp from hologram
        # FFT preprocessing
        # padding: 0 means no padding
        ["padding", fint, "Level of padding"],
        ["subtract mean", fbool, "Subtract mean before processing"],
        # pipeline_kws
        ["filter name", str, "Fourier filter used"],
        # qpretrieve defines the keyword argument `filter_size_interpretation`
        # for determining the filter size in Fourier space. In DC, we
        # need a well-defined value for the filter size. The most logical
        # choice is to interpret the filter size as "frequency index", which
        # is independent of the image shape and yields a good approximation
        # of the actual resolution one can expect. The default value
        # ("sideband distance") is a good choice for general QPI analysis,
        # but there is no meaningful information one could extract from it
        # by just looking at the number. Thus, the "filter size" that we
        # see here corresponds to a filter size set in qpretrieve where
        # `filter_size_interpretation="frequency index"`.
        ["filter size", float, "Fourier filter size [1/pix]"],
        ["scale to filter", fboolorfloat, "Scale QPI data to filter size"],
        # x, y coordinates, don't set if you wish None to be the default
        ["sideband freq", f1dfloatduple, "Sideband coordinates [1/pix]"],
        ["invert phase", fbool, "Invert the phase data"],
        # "pixel size proc" depends on `scale_to_filter`.
        # If `scale_to_filter` is False, this is equal to "pixel size raw".
        # If `scale_to_filter` is True or a float, this value will differ from
        # "pixel size raw".
        # RTDC "imaging:pixel size" equals "pixel size proc"
        ["pixel size proc", float, "QPI pixel size [µm]."],
        # postprocessing of phase and amplitude
        ["amp fit offset", str, "Amplitude offset correction"],
        ["amp fit profile", str, "Amplitude profile correction"],
        ["pha fit offset", str, "Phase offset correction"],
        ["pha fit profile", str, "Phase profile correction"],
        # QPImage background correction mask information
        ["amp border px", fint, "Width of border for amplitude [pix]"],
        ["pha border px", fint, "Width of border for phase [pix]"],
        # Forward compatible QPImage background correction for trivial
        # masks e.g. "tblr". "tb" useful for RTDC channel.
        ["amp border loc", str, "Border location specifier for amplitude"],
        ["pha border loc", str, "Border location specifier for phase"],
        # refocusing metadata
        ["focus interval", f1dfloatduple, "Focus interval to search [µm]"],
        ["focus metric", str, "Metric used to calculate focus"],
        ["focus minimizer", str, "Minimizer used to calculate focus"],
        ["focus kernel", str, "Propagation kernel"],
        ["focus padding", fint, "Level of padding for refocus"],
    ],
    # All setup-related keywords, except imaging
    "setup": [
        ["channel width", float, "Width of microfluidic channel [µm]"],
        ["chip identifier", lcstr, "Unique identifier of the chip used"],
        ["chip region", lcstr, "Imaged chip region (channel or reservoir)"],
        ["flow rate", float, "Flow rate in channel [µL/s]"],
        ["flow rate sample", float, "Sample flow rate [µL/s]"],
        ["flow rate sheath", float, "Sheath flow rate [µL/s]"],
        ["identifier", str, "Unique setup identifier"],
        # "medium" can have various values; it is used to calculate viscosity
        ["medium", str, "Medium used"],
        ["module composition", str, "Comma-separated list of modules used"],
        ["software version", str, "Acquisition software with version"],
        ["temperature", float, "Mean chip temperature [°C]"],
    ],
}

# CFG convenience lists and dicts
_cfg = copy.deepcopy(CFG_METADATA)
_cfg.update(CFG_ANALYSIS)

#: dict with metadata description
config_descr = {}
for _key in _cfg:
    config_descr[_key] = {}
    for _subkey, __, _descr in _cfg[_key]:
        config_descr[_key][_subkey] = _descr

#: dict of dicts containing functions to convert input data
config_funcs = {}
for _key in _cfg:
    config_funcs[_key] = {}
    for _subkey, _type, __ in _cfg[_key]:
        config_funcs[_key][_subkey] = _type

#: dict with section as keys and config parameter names as values
config_keys = {}
for _key in _cfg:
    config_keys[_key] = [it[0] for it in _cfg[_key]]

#: dict of dicts containing the type of section parameters
config_types = {}
for _key in _cfg:
    config_types[_key] = {}
    for _subkey, _type, __ in _cfg[_key]:
        if _type in func_types:
            _type = func_types[_type]
        config_types[_key][_subkey] = _type





class PipelineWarning(UserWarning):
    """Super-class for warnings relevant to data analysis

    There are those types of warnings in dclab that are
    important to the user, because they suggest that the
    user may not use the correct model (e.g. Young's modulus
    computation) in his analysis pipeline. All of these
    warnings should be subclassed from PipelineWarning
    to allow identifying them in higher-level software
    such as Shape-Out and to present them correctly to the
    user.
    """
    pass



#: List of scalar (one scalar value per event) features. This
#: list does not include the `ml_score_???` features. If you
#: need find out whether a feature name is valid, please use
#: :func:`is_valid_feature`.
FEATURES_SCALAR = [
    ["area_cvx", "Convex area [px]"],
    # area_msd is the contour moment M00
    ["area_msd", "Measured area [px]"],
    ["area_ratio", "Porosity (convex to measured area ratio)"],
    # area_um is computed from the convex contour
    ["area_um", "Area [µm²]"],
    ["area_um_raw", "Area [µm²] of raw contour"],
    ["aspect", "Aspect ratio of bounding box"],
    # The background brightness of the frame (not of the mask)
    ["bg_med", "Median frame background brightness [a.u.]"],
    # Background image offset which should be added to "image_bg" before
    # performing background correction. This is used in cases where the
    # background stored in "image_bg" is not accurate enough, e.g.
    # because the image_bg is a median image for multiple events and
    # the imaging system exhibits flickering.
    ["bg_off", "Background offset [a.u.]"],
    # Brightness values are computed only for pixels inside the mask
    ["bright_avg", "Brightness average [a.u.]"],
    ["bright_sd", "Brightness SD [a.u.]"],
    ["bright_bc_avg", "Brightness average (bgc) [a.u.]"],
    ["bright_bc_sd", "Brightness SD (bgc) [a.u.]"],
    ["bright_perc_10", "10th Percentile of brightness (bgc)"],
    ["bright_perc_90", "90th Percentile of brightness (bgc)"],
    ["circ", "Circularity"],
    # deform is computed from the convex contour
    ["deform", "Deformation"],
    ["deform_raw", "Deformation of raw contour"],
    ["eccentr_prnc", "Eccentricity of raw contour"],
    ["emodulus", "Young's modulus [kPa]"],
    # fl*_area, fl*_pos, and fl*_width values correspond to the
    # object for which the contour was found. For high concentrations,
    # these values could be error-prone due to the assignment from
    # false objects.
    ["fl1_area", "FL-1 area of peak [a.u.]"],
    # fl1_dist is set to zero if there is only one peak
    ["fl1_dist", "FL-1 distance between two first peaks [µs]"],
    ["fl1_max", "FL-1 maximum [a.u.]"],
    ["fl1_max_ctc", "FL-1 maximum, crosstalk-corrected [a.u.]"],
    ["fl1_npeaks", "FL-1 number of peaks"],
    ["fl1_pos", "FL-1 position of peak [µs]"],
    ["fl1_width", "FL-1 width [µs]"],
    ["fl2_area", "FL-2 area of peak [a.u.]"],
    ["fl2_dist", "FL-2 distance between two first peaks [µs]"],
    ["fl2_max", "FL-2 maximum [a.u.]"],
    ["fl2_max_ctc", "FL-2 maximum, crosstalk-corrected [a.u.]"],
    ["fl2_npeaks", "FL-2 number of peaks"],
    ["fl2_pos", "FL-2 position of peak [µs]"],
    ["fl2_width", "FL-2 width [µs]"],
    ["fl3_area", "FL-3 area of peak [a.u.]"],
    ["fl3_dist", "FL-3 distance between two first peaks [µs]"],
    ["fl3_max", "FL-3 maximum [a.u.]"],
    ["fl3_max_ctc", "FL-3 maximum, crosstalk-corrected [a.u.]"],
    ["fl3_npeaks", "FL-3 number of peaks"],
    ["fl3_pos", "FL-3 position of peak [µs]"],
    ["fl3_width", "FL-3 width [µs]"],
    # Sum of the flow rates for sample and sheath flow
    ["flow_rate", "Flow rate [µLs⁻¹]"],
    ["frame", "Video frame number"],
    ["g_force", "Gravitational force in multiples of g"],
    # index starts with 1
    ["index", "Index (Dataset)"],
    # index_online enumerates the events found by Shape-In and may have
    # missing values in case of a writer-overrun with recovery (#71)
    ["index_online", "Index (Online)"],
    # The inertia ratios of the event contours are defined by the
    # central second order moments of area (sqrt(m20/m02).
    ["inert_ratio_cvx", "Inertia ratio of convex contour"],
    ["inert_ratio_prnc", "Principal inertia ratio of raw contour"],
    ["inert_ratio_raw", "Inertia ratio of raw contour"],
    # This is an ancillary integer feature for visualizing the class
    # membership of individual events based on the `ml_score_???`
    # features.
    ["ml_class", "Most probable ML class"],
    ["nevents", "Number of events in the same image"],
    ["pc1", "Principal component 1"],
    ["pc2", "Principal component 2"],
    # Convexity is defined by the ratio of convex contour length
    # versus raw contour length. We use inverse convexity as it is
    # more convenient to work with numbers >1.
    ["per_ratio", "Inverse Convexity (raw to convex perimeter ratio)"],
    ["per_um_raw", "Perimeter [µm] of raw contour"],
    # pos_x and pos_y are computed from the contour moments
    # "m10"/"m00" and "m01"/"m00" of the convex hull of "contour"
    ["pos_x", "Position along channel axis [µm]"],
    ["pos_y", "Position lateral in channel [µm]"],
    # Sum of the pressures applied to sample and sheath flow
    ["pressure", "Pressure [mPa]"],
    # QPI features computed from holographic data
    ["qpi_dm_avg", "Dry mass (average) [pg]"],
    ["qpi_dm_sd", "Dry mass (SD) [pg]"],
    ["qpi_pha_int", "Integrated phase [rad]"],
    ["qpi_ri_avg", "Refractive index (average)"],
    ["qpi_ri_sd", "Refractive index (SD)"],
    # QPI features from refocused events
    ["qpi_focus", "Computed focus distance [µm]"],
    # Size features
    ["size_x", "Bounding box size x [µm]"],
    ["size_y", "Bounding box size y [µm]"],
    # Ratio between two halves of a mask
    ["sym_x", "Symmetry ratio left-right"],
    ["sym_y", "Symmetry ratio top-bottom"],
    ["temp", "Chip temperature [°C]"],
    ["temp_amb", "Ambient temperature [°C]"],
    # Haralick texture features can be computed using the mahotas package
    # from the background-corrected and masked image
    ["tex_asm_avg", "Texture angular second moment (avg)"],  # H1
    ["tex_asm_ptp", "Texture angular second moment (ptp)"],  # H1
    ["tex_con_avg", "Texture contrast (avg)"],  # H2
    ["tex_con_ptp", "Texture contrast (ptp)"],  # H2
    ["tex_cor_avg", "Texture correlation (avg)"],  # H3
    ["tex_cor_ptp", "Texture correlation (ptp)"],  # H3
    ["tex_den_avg", "Texture difference entropy (avg)"],  # 11
    ["tex_den_ptp", "Texture difference entropy (ptp)"],  # 11
    ["tex_ent_avg", "Texture entropy (avg)"],  # H9
    ["tex_ent_ptp", "Texture entropy (ptp)"],  # H9
    ["tex_f12_avg", "Texture First measure of correlation (avg)"],  # 12
    ["tex_f12_ptp", "Texture First measure of correlation (ptp)"],  # 12
    ["tex_f13_avg", "Texture Second measure of correlation (avg)"],  # 13
    ["tex_f13_ptp", "Texture Second measure of correlation (ptp)"],  # 13
    ["tex_idm_avg", "Texture inverse difference moment (avg)"],  # H5
    ["tex_idm_ptp", "Texture inverse difference moment (ptp)"],  # H5
    ["tex_sen_avg", "Texture sum entropy (avg)"],  # H8
    ["tex_sen_ptp", "Texture sum entropy (ptp)"],  # H8
    ["tex_sva_avg", "Texture sum variance (avg)"],  # H7
    ["tex_sva_ptp", "Texture sum variance (ptp)"],  # H7
    ["tex_var_avg", "Texture variance (avg)"],  # H4
    ["tex_var_ptp", "Texture variance (ptp)"],  # H4
    ["tilt", "Absolute tilt of raw contour"],
    ["time", "Time [s]"],
    # Volume is computed from the raw contour (i.e. with exclusions).
    # Fun fact: If we had decided to compute it from the convex contour,
    # then we would have close to none pixelation effects ¯\_(ツ)_/¯.
    ["volume", "Volume [µm³]"],
]

#: User-defined features: They are not reserved for anything specific
#: and can be used by the user for e.g. prototyping.
for _i in range(10):
    FEATURES_SCALAR.append([f"userdef{_i}", f"User-defined {_i}"])

#: Basin mapping features: These are used for datasets that are derived from a
#: subset of events of another dataset. For instance, if a dataset consists
#: of every second event from another dataset, the feature `basinmap1` would
#: consist of the integer array `[1, 3, 5, 7, ...]` (indexing starts at zero).
#: The `basinmap1` feature must then be referenced in the corresponding basin
#: definition. These features should not be presented explicitly to the
#: normal user (e.g. in Shape-Out) to avoid ambiguities, and they should
#: always be exported alongside basins that refer to them.
for _j in range(10):
    FEATURES_SCALAR.append([f"basinmap{_j}", f"Basin mapping {_j}"])

#: list of non-scalar features
FEATURES_NON_SCALAR = [
    # This is a (M, 2)-shaped array with integer contour coordinates
    ["contour", "Event contour"],
    ["image", "Gray scale event image"],
    ["image_bg", "Gray scale event background image"],
    # This is the contour with holes filled
    ["mask", "Binary mask labeling the event in the image"],
    # See FLUOR_TRACES for valid keys
    ["trace", "Dictionary of fluorescence traces"],
    # QPI experimental holographic data
    ["qpi_oah", "Off-axis hologram"],
    # QPI holographic background data (experimental or computed)
    ["qpi_oah_bg", "Off-axis hologram background"],
    # QPI features computed from holographic data
    ["qpi_pha", "Hologram phase image [rad]"],
    ["qpi_amp", "Hologram amplitude image"],
]

#: List of fluorescence traces
FLUOR_TRACES = [
    "fl1_median",
    "fl1_raw",
    "fl2_median",
    "fl2_raw",
    "fl3_median",
    "fl3_raw",
]


# FEATURE convenience lists and dicts

#: list of feature names
feature_names = [_cc[0] for _cc in FEATURES_SCALAR + FEATURES_NON_SCALAR]

#: list of feature labels (same order as :const:`feature_names`
feature_labels = [_cc[1] for _cc in FEATURES_SCALAR + FEATURES_NON_SCALAR]

#: dict for converting feature names to labels
feature_name2label = {}

for _cc in FEATURES_SCALAR + FEATURES_NON_SCALAR:
    feature_name2label[_cc[0]] = _cc[1]

#: list of scalar feature names
scalar_feature_names = [_cc[0] for _cc in FEATURES_SCALAR]



"""RT-DC hierarchy format"""
from __future__ import annotations

import numpy as np

from ... import definitions as dfn
from ...util import hashobj

from ..config import Configuration
from ..core import RTDCBase

from .events import (
    ChildContour, ChildNDArray, ChildScalar, ChildTrace, ChildTraceItem
)
from .hfilter import HierarchyFilter


class RTDC_Hierarchy(RTDCBase):
    def __init__(self, hparent, apply_filter=True, *args, **kwargs):
        """Hierarchy dataset (filtered from RTDCBase)

        A few words on hierarchies:
        The idea is that a subclass of RTDCBase can use the filtered
        data of another subclass of RTDCBase and interpret these data
        as unfiltered events. This comes in handy e.g. when the
        percentage of different subpopulations need to be distinguished
        without the noise in the original data.

        Children in hierarchies always update their data according to
        the filtered event data from their parent when `apply_filter`
        is called. This makes it easier to save and load hierarchy
        children with e.g. Shape-Out and it makes the handling of
        hierarchies more intuitive (when the parent changes,
        the child changes as well).

        Parameters
        ----------
        hparent: instance of RTDCBase
            The hierarchy parent
        apply_filter: bool
            Whether to apply the filter during instantiation;
            If set to `False`, `apply_filter` must be called
            manually.
        *args:
            Arguments for `RTDCBase`
        **kwargs:
            Keyword arguments for `RTDCBase`

        Attributes
        ----------
        hparent: RTDCBase
            Hierarchy parent of this instance
        """
        super(RTDC_Hierarchy, self).__init__(*args, **kwargs)

        self.path = hparent.path
        self.title = hparent.title + "_child"
        self._logs = None  # lazily-loaded
        self._tables = None  # lazily-loaded

        self._events = {}

        self.hparent = hparent

        self.config = self._create_config()  # init config
        self._update_config()  # sets e.g. event count

        if apply_filter:
            # Apply the filter
            # This will also populate all event attributes
            self.apply_filter()

        self._length = None

    def __contains__(self, key):
        return self.hparent.__contains__(key)

    def __getitem__(self, feat):
        """Return the feature data and cache them in self._events"""
        if feat in self._events:
            data = self._events[feat]
        elif feat in self.hparent:
            if len(self.hparent[feat].shape) > 1:
                # non-scalar feature
                data = ChildNDArray(self, feat)
            else:
                # scalar feature
                data = ChildScalar(self, feat)
            # Cache everything, even the Young's modulus. The user is
            # responsible for calling `rejuvenate` to reset everything.
            self._events[feat] = data
        else:
            raise KeyError(
                f"The dataset {self} does not contain the feature '{feat}'! "
                + "If you are attempting to access an ancillary feature "
                + "(e.g. emodulus), please make sure that the feature "
                + f"data are computed for {self.get_root_parent()} (the "
                + "root parent of this hierarchy child).")
        return data

    def __len__(self):
        if self._length is None:
            self._length = np.sum(self.hparent.filter.all)
        return self._length

    def _assert_filter(self):
        """Make sure filters exists

        Override from base class that uses :class:`.HierarchyFilter`
        instead of :class:`.Filter`.
        """
        if self._ds_filter is None:
            self._ds_filter = HierarchyFilter(self)

    def _check_parent_filter(self):
        """Reset filter if parent changed

        This will create a new HierarchyFilter for self if the
        parent RTDCBase changed. We do it like this, because it
        would be complicated to track all the changes in
        HierarchyFilter.
        """
        if self.filter.parent_changed:
            manual_pidx = self.filter.retrieve_manual_indices(self)
            self._ds_filter = None  # forces recreation of HierarchyFilter
            self._assert_filter()
            self.filter.apply_manual_indices(self, manual_pidx)

    def _create_config(self):
        """Return a stripped configuration from the parent"""
        # create a new configuration
        cfg = self.hparent.config.copy()
        # Remove previously applied filters
        pops = []
        for key in cfg["filtering"]:
            if (key.endswith(" min") or
                key.endswith(" max") or
                    key == "polygon filters"):
                pops.append(key)
        [cfg["filtering"].pop(key) for key in pops]
        # Add parent information in dictionary
        cfg["filtering"]["hierarchy parent"] = self.hparent.identifier
        return Configuration(cfg=cfg)

    def _update_config(self):
        """Update varying config values from self.hparent"""
        # event count
        self.config["experiment"]["event count"] = np.sum(
            self.hparent.filter.all)
        # calculation
        if "calculation" in self.hparent.config:
            self.config["calculation"].clear()
            self.config["calculation"].update(
                self.hparent.config["calculation"])

    @property
    def basins(self):
        return self.hparent.basins

    @property
    def features(self):
        return self.hparent.features

    @property
    def features_ancillary(self):
        return self.hparent.features_ancillary

    @property
    def features_basin(self):
        return self.hparent.features_basin

    @property
    def features_innate(self):
        return self.hparent.features_innate

    @property
    def features_loaded(self):
        return self.hparent.features_loaded

    @property
    def features_local(self):
        return self.hparent.features_local

    @property
    def features_scalar(self):
        return self.hparent.features_scalar

    @property
    def hash(self):
        """Hashes of a hierarchy child changes if the parent changes"""
        # Do not apply filters here (speed)
        hph = self.hparent.hash
        hpfilt = hashobj(self.hparent.filter.all)
        dhash = hashobj(hph + hpfilt)
        return dhash

    @property
    def logs(self):
        if self._logs is None:
            self._logs = self.hparent.logs
        return self._logs

    @logs.setter
    def logs(self, value):
        # to support setting `self.logs = {}` in RTDCBase
        if value:
            raise ValueError(
                "Setting actual logs not supported for RTDC_Hierarchy")

    @property
    def tables(self):
        if self._tables is None:
            self._tables = self.hparent.tables
        return self._tables

    @tables.setter
    def tables(self, value):
        # to support setting `self.tables = {}` in RTDCBase
        if value:
            raise ValueError(
                "Setting actual tables not supported for RTDC_Hierarchy")

    def apply_filter(self, *args, **kwargs):
        """Overridden `apply_filter` to perform tasks for hierarchy child"""
        if self._ds_filter is not None:
            # make sure self.filter knows about root manual indices
            # (stored in self.filter._man_root_ids)
            self.filter.retrieve_manual_indices(self)

        # Copy event data from hierarchy parent
        self.hparent.apply_filter(*args, **kwargs)

        # Clear anything that has been cached until now
        self._length = None

        # update event index
        event_count = len(self)
        self._events.clear()
        self._events["index"] = np.arange(1, event_count + 1)
        # set non-scalar column data
        for feat in ["image", "image_bg", "mask"]:
            if feat in self.hparent:
                self._events[feat] = ChildNDArray(self, feat)
        if "contour" in self.hparent:
            self._events["contour"] = ChildContour(self)
        if "trace" in self.hparent:
            trdict = ChildTrace()
            for flname in dfn.FLUOR_TRACES:
                if flname in self.hparent["trace"]:
                    trdict[flname] = ChildTraceItem(self, flname)
            self._events["trace"] = trdict
        # Update configuration
        self._update_config()

        # create a new filter if the parent changed
        self._check_parent_filter()
        super(RTDC_Hierarchy, self).apply_filter(*args, **kwargs)

    def get_root_parent(self):
        """Return the root parent of this dataset"""
        if isinstance(self.hparent, RTDC_Hierarchy):
            return self.hparent.get_root_parent()
        else:
            return self.hparent

    def rejuvenate(self):
        """Redraw the hierarchy tree, updating config and features

        You should call this function whenever you change something
        in the hierarchy parent(s), be it filters or metadata for computing
        ancillary features.

        .. versionadded: 0.47.0
            This is just an alias of `apply_filter`, but with a more
            accurate name (not only the filters are applied, but alot
            of other things might change).
        """
        self.apply_filter()



"""RT-DC dataset core classes and methods"""

import warnings

import numpy as np

from dclab import definitions as dfn

from .. import downsampling
from ..polygon_filter import PolygonFilter


class NanWarning(UserWarning):
    pass


class Filter(object):
    def __init__(self, rtdc_ds):
        """Boolean filter arrays for RT-DC measurements

        Parameters
        ----------
        rtdc_ds: instance of RTDCBase
            The RT-DC dataset the filter applies to
        """
        # dictionary of boolean array for box filters
        self._box_filters = {}
        # dictionary of (hash, boolean array) for polygon filters
        self._poly_filters = {}
        # dictionary of all internal property filters
        self._array_props = {}
        # initialize important parameters
        self._init_rtdc_ds(rtdc_ds)
        # initialize properties
        self.reset()

    def __getitem__(self, key):
        """Return the filter for a feature in `self.features`"""
        if key in self.features and dfn.scalar_feature_exists(key):
            if key not in self._box_filters:
                # Generate filters on-the-fly
                self._box_filters[key] = np.ones(self.size, dtype=bool)
        else:
            raise KeyError("Feature not available: '{}'".format(key))
        return self._box_filters[key]

    @property
    def all(self):
        """All filters combined (see :func:`Filter.update`)

        Use this property to filter the features of
        :class:`dclab.rtdc_dataset.RTDCBase` instances
        """
        return self._get_ro_array("all")

    @property
    def box(self):
        """All box filters"""
        return self._get_ro_array("box")

    @property
    def polygon(self):
        """Polygon filters"""
        return self._get_ro_array("polygon")

    @property
    def invalid(self):
        """Invalid (nan/inf) events"""
        return self._get_ro_array("invalid")

    def _get_ro_array(self, key):
        view = self._get_rw_array(key).view()
        view.flags.writeable = False
        return view

    def _get_rw_array(self, key):
        if key not in self._array_props:
            self._array_props[key] = np.ones(self.size, dtype=bool)
        return self._array_props[key]

    def _init_rtdc_ds(self, rtdc_ds):
        #: Available feature names
        self.features = rtdc_ds.features_scalar
        if hasattr(self, "size") and self.size != len(rtdc_ds):
            raise ValueError("Change of RTDCBase size not supported!")
        self.size = len(rtdc_ds)
        # determine box filters that have been removed
        for key in list(self._box_filters.keys()):
            if key not in self.features:
                self._box_filters.pop(key)
        # determine polygon filters that have been removed
        for pf_id in list(self._poly_filters.keys()):
            pf = PolygonFilter.get_instance_from_id(pf_id)
            if (pf_id in rtdc_ds.config["filtering"]["polygon filters"]
                and pf.axes[0] in self.features
                    and pf.axes[1] in self.features):
                pass
            else:
                # filter has been removed
                self._poly_filters.pop(pf_id)

    def reset(self):
        """Reset all filters"""
        self._box_filters.clear()
        self._poly_filters.clear()
        self._array_props.clear()
        #: 1D boolean array for manually excluding events; `False` values
        #: are excluded.
        self.manual = np.ones(self.size, dtype=bool)
        # old filter configuration of `rtdc_ds`
        self._old_config = {}

    def update(self, rtdc_ds, force=None):
        """Update the filters according to `rtdc_ds.config["filtering"]`

        Parameters
        ----------
        rtdc_ds: dclab.rtdc_dataset.core.RTDCBase
            The measurement to which the filter is applied
        force : list
            A list of feature names that must be refiltered with
            min/max values.

        Notes
        -----
        This function is called when
        :func:`ds.apply_filter <dclab.rtdc_dataset.RTDCBase.apply_filter>`
        is called.
        """
        if force is None:
            force = []
        # re-initialize important parameters
        self._init_rtdc_ds(rtdc_ds)

        # These lists may help us become very fast in the future
        newkeys = []
        oldvals = []
        newvals = []

        cfg_cur = rtdc_ds.config["filtering"]
        cfg_old = self._old_config

        # Determine which data was updated
        for skey in list(cfg_cur.keys()):
            if cfg_cur[skey] != cfg_old.get(skey, None):
                newkeys.append(skey)
                oldvals.append(cfg_old.get(skey, None))
                newvals.append(cfg_cur[skey])

        # 1. Invalid filters
        arr_invalid = self._get_rw_array("invalid")
        arr_invalid[:] = True
        if cfg_cur["remove invalid events"]:
            for feat in self.features:
                data = rtdc_ds[feat]
                invalid = np.isinf(data) | np.isnan(data)
                arr_invalid &= ~invalid

        # 2. Filter all feature min/max values.
        feat2filter = []
        for k in newkeys:
            # k[:-4] because we want to crop " min" and " max"
            if (dfn.scalar_feature_exists(k[:-4])
                    and (k.endswith(" min") or k.endswith(" max"))):
                feat2filter.append(k[:-4])

        for f in force:
            # add forced features
            if dfn.scalar_feature_exists(f):
                feat2filter.append(f)
            else:
                # Make sure the feature name is valid.
                raise ValueError("Unknown scalar feature name '{}'!".format(f))

        feat2filter = np.unique(feat2filter)

        for feat in feat2filter:
            fstart = feat + " min"
            fend = feat + " max"
            must_be_filtered = (fstart in cfg_cur
                                and fend in cfg_cur
                                and cfg_cur[fstart] != cfg_cur[fend])
            if ((fstart in cfg_cur and fend not in cfg_cur)
                    or (fstart not in cfg_cur and fend in cfg_cur)):
                # User is responsible for setting min and max values!
                raise ValueError("Box filter: Please make sure that both "
                                 "'{}' and '{}' are set!".format(fstart, fend))
            if feat in self.features:
                # Get the current feature filter
                feat_filt = self[feat]
                feat_filt[:] = True
                # If min and max exist and if they are not identical:
                if must_be_filtered:
                    ivalstart = cfg_cur[fstart]
                    ivalend = cfg_cur[fend]
                    if ivalstart > ivalend:
                        msg = "inverting filter: {} > {}".format(fstart, fend)
                        warnings.warn(msg)
                        ivalstart, ivalend = ivalend, ivalstart
                    data = rtdc_ds[feat]
                    # treat nan-values in a special way
                    disnan = np.isnan(data)
                    if np.sum(disnan):
                        # this avoids RuntimeWarnings (invalid value
                        # encountered due to nan-values)
                        feat_filt[disnan] = False
                        idx = ~disnan
                        if not cfg_cur["remove invalid events"]:
                            msg = "Feature '{}' contains ".format(feat) \
                                  + "nan-values! Box filters remove those."
                            warnings.warn(msg, NanWarning)
                    else:
                        idx = slice(0, self.size)  # place-holder for [:]
                    feat_filt[idx] &= ivalstart <= data[idx]
                    feat_filt[idx] &= data[idx] <= ivalend
            elif must_be_filtered:
                warnings.warn("Dataset '{}' does ".format(rtdc_ds.identifier)
                              + "not contain the feature '{}'! ".format(feat)
                              + "A box filter has been ignored.")
        # store box filters
        arr_box = self._get_rw_array("box")
        arr_box[:] = True
        for feat in self._box_filters:
            arr_box &= self._box_filters[feat]

        # 3. Filter with polygon filters
        # check if something has changed
        # perform polygon filtering
        for pf_id in cfg_cur["polygon filters"]:
            pf = PolygonFilter.get_instance_from_id(pf_id)
            if (pf_id not in self._poly_filters
                    or pf.hash != self._poly_filters[pf_id][0]):
                datax = rtdc_ds[pf.axes[0]]
                datay = rtdc_ds[pf.axes[1]]
                self._poly_filters[pf_id] = (pf.hash, pf.filter(datax, datay))
        # store polygon filters
        arr_polygon = self._get_rw_array("polygon")
        arr_polygon[:] = True
        for pf_id in self._poly_filters:
            arr_polygon &= self._poly_filters[pf_id][1]

        # 4. Finally combine all filters and apply "limit events"
        # get a list of all filters
        arr_all = self._get_rw_array("all")
        if cfg_cur["enable filters"]:
            arr_all[:] = arr_box & arr_invalid & arr_polygon & self.manual

            # Filter with configuration keyword argument "limit events".
            # This additional step limits the total number of events in
            # self.all.
            if cfg_cur["limit events"] > 0:
                limit = cfg_cur["limit events"]
                sub = arr_all[arr_all]
                _, idx = downsampling.downsample_rand(sub,
                                                      samples=limit,
                                                      ret_idx=True)
                sub[~idx] = False
                arr_all[arr_all] = sub
        else:
            arr_all[:] = True

        # Actual filtering is then done during plotting
        self._old_config = rtdc_ds.config.copy()["filtering"]



import functools
# import multiprocessing BaseManager here, because there is some kind
# of circular dependency issue with s3transfer.compat and multiprocessing.
from multiprocessing.managers import BaseManager  # noqa: F401
import os
import pathlib
import re
import socket
from urllib.parse import urlparse
import warnings


try:
    import boto3
    import botocore
    import botocore.client
    import botocore.exceptions
    import botocore.session
except ModuleNotFoundError:
    BOTO3_AVAILABLE = False
else:
    BOTO3_AVAILABLE = True

from ..http_utils import HTTPFile

from .feat_basin import Basin

from .fmt_hdf5 import RTDC_HDF5


#: Regular expression for matching a DCOR resource URL
REGEXP_S3_URL = re.compile(
    r"^(https?:\/\/)"  # protocol (http or https or omitted)
    r"([a-z0-9-\.]*)(\:[0-9]*)?\/"  # host:port
    r".+\/"  # bucket
    r".+"  # key
)
REGEXP_S3_BUCKET_KEY = re.compile(r"^[0-9a-z-]+(\/[0-9a-z-]+)+$")

S3_ENDPOINT_URL = os.environ.get("DCLAB_S3_ENDPOINT_URL")
S3_ACCESS_KEY_ID = os.environ.get("DCLAB_S3_ACCESS_KEY_ID")
S3_SECRET_ACCESS_KEY = os.environ.get("DCLAB_S3_SECRET_ACCESS_KEY")


class S3File(HTTPFile):
    """Monkeypatched `HTTPFile` to support authenticated access to S3"""
    def __init__(self,
                 object_path: str,
                 endpoint_url: str,
                 access_key_id: str = "",
                 secret_access_key: str = "",
                 use_ssl: bool = True,
                 verify_ssl: bool = True):
        """

        Parameters
        ----------
        object_path: str
            bucket/key path to object in the object store
        endpoint_url: str
            the explicit endpoint URL for accessing the object store
        access_key_id:
            S3 access key
        secret_access_key:
            secret S3 key mathcing `access_key_id`
        use_ssl: bool
            use SSL to connect to the endpoint, only disabled for testing
        verify_ssl: bool
            make sure the SSL certificate is sound, only used for testing
        """
        if endpoint_url is None:
            raise ValueError(
                "The S3 endpoint URL is empty. This could mean that you did "
                "not specify the full S3 URL or that you forgot to set "
                "the `S3_ENDPOINT_URL` environment variable.")
        endpoint_url = endpoint_url.strip().rstrip("/")
        self.botocore_session = botocore.session.get_session()
        self.s3_session = boto3.Session(
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
            botocore_session=self.botocore_session)
        self.s3_client = self.s3_session.client(
            service_name='s3',
            use_ssl=use_ssl,
            verify=verify_ssl,
            endpoint_url=endpoint_url,
            )
        # Use a configuration that allows anonymous access
        # https://stackoverflow.com/a/34866092
        if not secret_access_key:
            config = botocore.client.Config(
                signature_version=botocore.UNSIGNED,
                region_name='us-east-1')
        else:
            config = None

        self.s3_resource = self.s3_session.resource(
            service_name="s3",
            use_ssl=use_ssl,
            verify=verify_ssl,
            endpoint_url=endpoint_url,
            config=config)

        bucket_name, object_name = object_path.strip("/").split("/", 1)
        self.s3_object = self.s3_resource.Object(
            bucket_name=bucket_name,
            key=object_name)

        super(S3File, self).__init__(f"{endpoint_url}/{object_path}")

    def _parse_header(self):
        if self._len is None:
            self._len = self.s3_object.content_length
            self._etag = self.s3_object.e_tag

    def close(self):
        super(S3File, self).close()
        self.s3_client.close()

    def download_range(self, start, stop):
        """Download bytes given by the range (`start`, `stop`)

        `stop` is not inclusive (In the HTTP range request it normally is).
        """
        stream = self.s3_object.get(Range=f"bytes={start}-{stop-1}")['Body']
        return stream.read()


class RTDC_S3(RTDC_HDF5):
    def __init__(self,
                 url: str,
                 endpoint_url: str = None,
                 access_key_id: str = None,
                 secret_access_key: str = None,
                 use_ssl: bool = True,
                 *args, **kwargs):
        """Access RT-DC measurements in an S3-compatible object store

        This is essentially just a wrapper around :class:`.RTDC_HDF5`
        with :mod:`boto3` and :class:`.HTTPFile` passing a file object to h5py.

        Parameters
        ----------
        url: str
            URL to an object in an S3 instance; this can be either a full
            URL (including the endpoint), or just `bucket/key`
        access_key_id: str
            S3 access identifier
        secret_access_key: str
            Secret S3 access key
        use_ssl: bool
            Whether to enforce SSL (defaults to True)
        *args:
            Arguments for `RTDCBase`
        **kwargs:
            Keyword arguments for `RTDCBase`

        Attributes
        ----------
        path: str
            The URL to the object
        """
        if not BOTO3_AVAILABLE:
            raise ModuleNotFoundError(
                f"Package `boto3` required for loading S3 data '{url}'!")

        self._s3file = S3File(
            object_path=get_object_path(url),
            endpoint_url=(endpoint_url
                          or get_endpoint_url(url)
                          or S3_ENDPOINT_URL),
            access_key_id=(access_key_id
                           or S3_ACCESS_KEY_ID
                           or ""),
            secret_access_key=(secret_access_key
                               or S3_SECRET_ACCESS_KEY
                               or ""),
            use_ssl=use_ssl,
            verify_ssl=use_ssl,
            )
        # Initialize the HDF5 dataset
        super(RTDC_S3, self).__init__(
            h5path=self._s3file,
            *args,
            **kwargs)
        # Override self.path with the actual S3 URL
        self.path = self._s3file.url

    def close(self):
        super(RTDC_S3, self).close()
        self._s3file.close()


class S3Basin(Basin):
    basin_format = "s3"
    basin_type = "remote"

    def __init__(self, *args, **kwargs):
        self._available_verified = None
        super(S3Basin, self).__init__(*args, **kwargs)

    def _load_dataset(self, location, **kwargs):
        h5file = RTDC_S3(location, **kwargs)
        return h5file

    def is_available(self):
        """Check for boto3 and object availability

        Caching policy: Once this method returns True, it will always
        return True.
        """
        if self._available_verified is None:
            with self._av_check_lock:
                if not BOTO3_AVAILABLE:
                    self._available_verified = False
                else:
                    self._available_verified = \
                            is_s3_object_available(self.location)
        return self._available_verified


def is_s3_object_available(url: str,
                           access_key_id: str = None,
                           secret_access_key: str = None,
                           ):
    """Check whether an S3 object is available

    Parameters
    ----------
    url: str
        full URL to the object
    access_key_id: str
        S3 access identifier
    secret_access_key: str
        Secret S3 access key
    """
    avail = False
    if is_s3_url(url):
        endpoint_url = get_endpoint_url(url) or S3_ENDPOINT_URL
        if not endpoint_url:
            warnings.warn(
                f"Could not determine endpoint from URL '{url}'. Please "
                f"set the `S3_ENDPOINT_URL` environment variable or pass "
                f"a full object URL.")
        else:
            # default to https if no scheme or port is specified
            urlp = urlparse(endpoint_url)
            port = urlp.port or (80 if urlp.scheme == "http" else 443)
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.settimeout(1)
                # Try to connect to the host
                try:
                    # Use `hostname`, not `netloc`, because `netloc` contains
                    # the port number which we do not want here.
                    s.connect((urlp.hostname, port))
                except (socket.gaierror, OSError):
                    pass
                else:
                    # Try to access the object
                    s3file = S3File(
                        object_path=get_object_path(url),
                        endpoint_url=endpoint_url,
                        access_key_id=(access_key_id
                                       or S3_ACCESS_KEY_ID
                                       or ""),
                        secret_access_key=(secret_access_key
                                           or S3_SECRET_ACCESS_KEY
                                           or ""),
                        )
                    try:
                        s3file.s3_object.load()
                    except botocore.exceptions.ClientError:
                        avail = False
                    else:
                        avail = True
    return avail


@functools.lru_cache()
def get_endpoint_url(url):
    """Given a URL of an S3 object, return the endpoint URL

    Return None if no endpoint URL can be extracted (e.g. because
    just `bucket_name/object_path` was passed).
    """
    urlp = urlparse(url=url)
    if urlp.hostname:
        scheme = urlp.scheme or "https"
        port = urlp.port or (80 if scheme == "http" else 443)
        return f"{scheme}://{urlp.hostname}:{port}"
    else:
        return None


@functools.lru_cache()
def get_object_path(url):
    """Given a URL of an S3 object, return the `bucket_name/object_path` part

    Return object paths always without leading slash `/`.
    """
    urlp = urlparse(url=url)
    return urlp.path.lstrip("/")


@functools.lru_cache()
def is_s3_url(string):
    """Check whether `string` is a valid S3 URL using regexp"""
    if not isinstance(string, str):
        return False
    elif REGEXP_S3_URL.match(string.strip()):
        # this is pretty clear
        return True
    elif pathlib.Path(string).exists():
        # this is actually a file
        return False
    elif REGEXP_S3_BUCKET_KEY.match(string.strip()):
        # bucket_name/key
        return True
    else:
        return False



"""DCOR-med access token (SSL certificate + CKAN token)"""
import pathlib
import ssl
import tempfile
import zipfile


def get_api_key(access_token_path, password):
    """Extract the API key / API token from an encrypted DCOR access token"""
    if isinstance(password, str):
        password = password.encode("utf-8")
    with zipfile.ZipFile(access_token_path) as arc:
        api_key = arc.read("api_key.txt", pwd=password).decode().strip()
    return api_key


def get_certificate(access_token_path, password):
    """Extract the certificate bundle from an encrypted DCOR access token"""
    if isinstance(password, str):
        password = password.encode("utf-8")
    with zipfile.ZipFile(access_token_path) as arc:
        cert_data = arc.read("server.cert", pwd=password)
    return cert_data


def get_hostname(access_token_path, password):
    """Extract the hostname from an encrypted DCOR access token"""
    cert_data = get_certificate(access_token_path, password)
    with tempfile.TemporaryDirectory(prefix="dcoraid_access_token_") as td:
        cfile = pathlib.Path(td) / "server.cert"
        cfile.write_bytes(cert_data)
        # Dear future-self,
        #
        # I know that this will probably not have been a good solution.
        # Anyway, I still decided to use this private function from the
        # built-in ssh module to avoid additional dependencies. Just so
        # you know: If you happen to be in trouble now because of this,
        # bear in mind that you had enough time to at least ask for the
        # functionality to be implemented in the requests library. Look
        # how I kept the lines all the same length!
        #
        # Cheers,
        # Paul
        cert_dict = ssl._ssl._test_decode_cert(str(cfile))
    # get the common name
    for ((key, value),) in cert_dict["subject"]:
        if key == "commonName":
            hostname = value.strip()
            break
    else:
        raise KeyError("Could not extract hostname from certificate!")
    return hostname



"""
.. versionadded:: 0.33.0
"""
from __future__ import annotations

from typing import Optional

import numpy as np

from ..definitions import feat_logic

from .core import RTDCBase
from .fmt_hierarchy import RTDC_Hierarchy, map_indices_child2root


_registered_temporary_features = []


def deregister_all():
    """Deregisters all temporary features"""
    for feat in list(_registered_temporary_features):
        deregister_temporary_feature(feat)


def deregister_temporary_feature(feature: str):
    """Convenience function for deregistering a temporary feature

    This method is mostly used during testing. It does not
    remove the actual feature data from any dataset; the data
    will stay in memory but is not accessible anymore through
    the public methods of the :class:`RTDCBase` user interface.
    """
    if feature in _registered_temporary_features:
        _registered_temporary_features.remove(feature)
        feat_logic.feature_deregister(feature)


def register_temporary_feature(feature: str,
                               label: Optional[str] = None,
                               is_scalar: bool = True):
    """Register a new temporary feature

    Temporary features are custom features that can be defined ad hoc
    by the user. Temporary features are helpful when the integral
    features are not enough, e.g. for prototyping, testing, or
    collating with other data. Temporary features allow you to
    leverage the full functionality of :class:`RTDCBase` with
    your custom features (no need to go for a custom `pandas.Dataframe`).

    Parameters
    ----------
    feature: str
        Feature name; allowed characters are lower-case letters,
        digits, and underscores
    label: str
        Feature label used e.g. for plotting
    is_scalar: bool
        Whether or not the feature is a scalar feature
    """
    feat_logic.feature_register(feature, label, is_scalar)
    _registered_temporary_features.append(feature)


def set_temporary_feature(rtdc_ds: RTDCBase,
                          feature: str,
                          data: np.ndarray):
    """Set temporary feature data for a dataset

    Parameters
    ----------
    rtdc_ds: dclab.RTDCBase
        Dataset for which to set the feature. Note that the
        length of the feature `data` must match the number of events
        in `rtdc_ds`. If the dataset is a hierarchy child, the data will also
        be set in the parent dataset, but only for those events that are part
        of the child. For all events in the parent dataset that are not part
        of the child dataset, the temporary feature is set to np.nan.
    feature: str
        Feature name
    data: np.ndarray
        The data
    """
    if not feat_logic.feature_exists(feature):
        raise ValueError(
            f"Temporary feature '{feature}' has not been registered!")
    if len(data) != len(rtdc_ds):
        raise ValueError(f"The temporary feature {feature} must have same "
                         f"length as the dataset. Expected length "
                         f"{len(rtdc_ds)}, got length {len(data)}!")
    if isinstance(rtdc_ds, RTDC_Hierarchy):
        root_ids = map_indices_child2root(rtdc_ds, np.arange(len(rtdc_ds)))
        root_parent = rtdc_ds.get_root_parent()
        root_feat_data = np.empty((len(root_parent)))
        root_feat_data[:] = np.nan
        root_feat_data[root_ids] = data
        set_temporary_feature(root_parent, feature, root_feat_data)
        rtdc_ds.rejuvenate()
    else:
        feat_logic.check_feature_shape(feature, data)
        data_ro = data.view()
        data_ro.setflags(write=False)
        rtdc_ds._usertemp[feature] = data_ro
